from typing import Callable, Optional

import torch
import torch.nn as nn

from tensordict import TensorDict
from torch import Tensor

from rl4co.models.common.constructive.nonautoregressive import NonAutoregressiveEncoder
from rl4co.models.nn.env_embeddings import env_edge_embedding, env_init_embedding
from rl4co.models.nn.graph.gnn import GNNEncoder

try:
    from torch_geometric.data import Batch
except ImportError:
    # `Batch` is referred to only as type notations in this file
    Batch = None


class EdgeHeatmapGenerator(nn.Module):
    """MLP for converting edge embeddings to heatmaps.

    Args:
        embed_dim: Dimension of the embeddings
        num_layers: The number of linear layers in the network.
        act_fn: Activation function. Defaults to "silu".
        linear_bias: Use bias in linear layers. Defaults to True.
        undirected_graph: Whether the graph is undirected. Defaults to True.
    """

    def __init__(
        self,
        embed_dim: int,
        num_layers: int,
        act_fn: str | Callable = "silu",
        linear_bias: bool = True,
        undirected_graph: bool = True,
    ) -> None:
        super(EdgeHeatmapGenerator, self).__init__()

        self.linears = nn.ModuleList(
            [
                nn.Linear(embed_dim, embed_dim, bias=linear_bias)
                for _ in range(num_layers - 1)
            ]
        )
        self.output = nn.Linear(embed_dim, 1, bias=linear_bias)

        self.act = getattr(nn.functional, act_fn) if isinstance(act_fn, str) else act_fn

        self.undirected_graph = undirected_graph

    def forward(self, graph: Batch) -> Tensor:  # type: ignore
        # do not reuse the input value
        edge_attr = graph.edge_attr  # type: ignore
        for layer in self.linears:
            edge_attr = self.act(layer(edge_attr))
        graph.edge_attr = torch.sigmoid(self.output(edge_attr))  # type: ignore

        heatmap_logits = self._make_heatmap_logits(graph)
        return heatmap_logits

    def _make_heatmap_logits(self, batch_graph: Batch) -> Tensor:  # type: ignore
        graphs = batch_graph.to_data_list()
        device = graphs[0].edge_attr.device
        batch_size = len(graphs)
        num_nodes = graphs[0].x.shape[0]

        heatmap = torch.zeros(
            (batch_size, num_nodes, num_nodes),
            device=device,
            dtype=graphs[0].edge_attr.dtype,
        )

        for index, graph in enumerate(graphs):
            edge_index, edge_attr = graph.edge_index, graph.edge_attr
            heatmap[index, edge_index[0], edge_index[1]] = edge_attr.flatten()

        # This is commented out, because it undo the some of the sparsification.
        # if self.undirected_graph:
        #     heatmap = (heatmap + heatmap.transpose(1, 2)) * 0.5

        # Avoid log(0) by adding a small value
        if heatmap.dtype == torch.float32 or heatmap.dtype == torch.bfloat16:
            small_value = 1e-12
        elif heatmap.dtype == torch.float16:
            small_value = 3e-8  # the smallest positive number such that log(small_value) is not -inf
        else:
            raise ValueError(f"Unsupported dtype: {heatmap.dtype}")

        heatmap += small_value
        heatmap_logits = torch.log(heatmap)

        return heatmap_logits


class NARGNNEncoder(NonAutoregressiveEncoder):
    """Anisotropic Graph Neural Network encoder with edge-gating mechanism as in Joshi et al. (2022), and used in DeepACO (Ye et al., 2023).
    This creates a heatmap of NxN for N nodes (i.e., heuristic) that models the probability to go from one node to another for all nodes.
    This model utilizes a multi-layer perceptron (MLP) approach to predict edge attributes directly from the input graph features,
    which are then transformed into a heatmap representation to facilitate the decoding of the solution. The decoding process
    is managed by a specified strategy which could vary from simple greedy selection to more complex sampling methods.

    Tip:
        This decoder's performance heavily relies on the ability of the MLP to capture the dependencies between different
        parts of the solution without the iterative refinement provided by autoregressive models. It is particularly useful
        in scenarios where the solution space can be effectively explored in a parallelized manner or when the solution components
        are largely independent.

    Args:
        embed_dim: Dimension of the node embeddings
        env_name: Name of the environment used to initialize embeddings
        num_layers: Number of layers in the encoder
        init_embedding: Model to use for the initial embedding. If None, use the default embedding for the environment
        edge_embedding: Model to use for the edge embedding. If None, use the default embedding for the environment
        graph_network: Model to use for the graph network. If None, use the default network for the environment
        heatmap_generator: Model to use for the heatmap generator. If None, use the default network for the environment
        num_layers_heatmap_generator: Number of layers in the heatmap generator
        num_layers_graph_encoder: Number of layers in the graph encoder
        act_fn: The activation function to use in each GNNLayer, see https://pytorch.org/docs/stable/nn.functional.html#non-linear-activation-functions for available options. Defaults to 'silu'.
        agg_fn: The aggregation function to use in each GNNLayer for pooling features. Options: 'add', 'mean', 'max'. Defaults to 'mean'.
        linear_bias: Use bias in linear layers. Defaults to True.
        k_sparse: Number of edges to keep for each node. Defaults to None.
    """

    def __init__(
        self,
        embed_dim: int = 64,
        env_name: str = "tsp",
        # TODO: pass network
        init_embedding: Optional[nn.Module] = None,
        edge_embedding: Optional[nn.Module] = None,
        graph_network: Optional[nn.Module] = None,
        heatmap_generator: Optional[nn.Module] = None,
        num_layers_heatmap_generator: int = 5,
        num_layers_graph_encoder: int = 15,
        act_fn="silu",
        agg_fn="mean",
        linear_bias: bool = True,
        k_sparse: Optional[int] = None,
    ):
        super(NonAutoregressiveEncoder, self).__init__()
        self.env_name = env_name

        self.init_embedding = (
            env_init_embedding(self.env_name, {"embed_dim": embed_dim})
            if init_embedding is None
            else init_embedding
        )

        self.edge_embedding = (
            env_edge_embedding(
                self.env_name, {"embed_dim": embed_dim, "k_sparse": k_sparse}
            )
            if edge_embedding is None
            else edge_embedding
        )

        self.graph_network = (
            GNNEncoder(
                embed_dim=embed_dim,
                num_layers=num_layers_graph_encoder,
                act_fn=act_fn,
                agg_fn=agg_fn,
            )
            if graph_network is None
            else graph_network
        )

        self.heatmap_generator = (
            EdgeHeatmapGenerator(
                embed_dim=embed_dim,
                num_layers=num_layers_heatmap_generator,
                linear_bias=linear_bias,
            )
            if heatmap_generator is None
            else heatmap_generator
        )

    def forward(self, td: TensorDict):
        """Forward pass of the encoder.
        Transform the input TensorDict into the latent representation.
        """
        # Transfer to embedding space
        node_embed = self.init_embedding(td)
        graph = self.edge_embedding(td, node_embed)

        # Process embedding into graph
        # TODO: standardize?
        graph.x, graph.edge_attr = self.graph_network(
            graph.x, graph.edge_index, graph.edge_attr
        )

        # Generate heatmap logits
        heatmap_logits = self.heatmap_generator(graph)

        # Return latent representation (i.e. heatmap logits) and initial embeddings
        return heatmap_logits, node_embed


class NARGNNNodeEncoder(NARGNNEncoder):
    """In this case, we just use the node embeddings from the graph
    without transforming them into a heatmap.
    """

    def forward(self, td: TensorDict):
        # Transfer to embedding space
        node_embed = self.init_embedding(td)
        graph = self.edge_embedding(td, node_embed)

        # Process embedding into graph
        # TODO: standardize?
        graph.x, graph.edge_attr = self.graph_network(
            graph.x, graph.edge_index, graph.edge_attr
        )

        proc_embeds = graph.x
        batch_size = node_embed.shape[0]
        # reshape proc_embeds from [bs*n, h] to [bs, n, h]
        proc_embeds = proc_embeds.reshape(batch_size, -1, proc_embeds.shape[1])
        return proc_embeds, node_embed



import torch
import torch.nn as nn

from tensordict import TensorDict

from rl4co.utils.ops import batched_scatter_sum, gather_by_index


def env_context_embedding(env_name: str, config: dict) -> nn.Module:
    """Get environment context embedding. The context embedding is used to modify the
    query embedding of the problem node of the current partial solution.
    Usually consists of a projection of gathered node embeddings and features to the embedding space.

    Args:
        env: Environment or its name.
        config: A dictionary of configuration options for the environment.
    """
    embedding_registry = {
        "tsp": TSPContext,
        "atsp": TSPContext,
        "cvrp": VRPContext,
        "cvrptw": VRPTWContext,
        "cvrpmvc": VRPContext,
        "ffsp": FFSPContext,
        "svrp": SVRPContext,
        "sdvrp": VRPContext,
        "pctsp": PCTSPContext,
        "spctsp": PCTSPContext,
        "op": OPContext,
        "dpp": DPPContext,
        "mdpp": DPPContext,
        "pdp": PDPContext,
        "mdcpdp": MDCPDPContext,
        "mtsp": MTSPContext,
        "smtwtp": SMTWTPContext,
        "mtvrp": MTVRPContext,
        "shpp": TSPContext,
        "flp": FLPContext,
        "mcp": MCPContext,
    }

    if env_name not in embedding_registry:
        raise ValueError(
            f"Unknown environment name '{env_name}'. Available context embeddings: {embedding_registry.keys()}"
        )

    return embedding_registry[env_name](**config)


class EnvContext(nn.Module):
    """Base class for environment context embeddings. The context embedding is used to modify the
    query embedding of the problem node of the current partial solution.
    Consists of a linear layer that projects the node features to the embedding space."""

    def __init__(self, embed_dim, step_context_dim=None, linear_bias=False):
        super(EnvContext, self).__init__()
        self.embed_dim = embed_dim
        step_context_dim = step_context_dim if step_context_dim is not None else embed_dim
        self.project_context = nn.Linear(step_context_dim, embed_dim, bias=linear_bias)

    def _cur_node_embedding(self, embeddings, td):
        """Get embedding of current node"""
        cur_node_embedding = gather_by_index(embeddings, td["current_node"])
        return cur_node_embedding

    def _state_embedding(self, embeddings, td):
        """Get state embedding"""
        raise NotImplementedError("Implement for each environment")

    def forward(self, embeddings, td):
        cur_node_embedding = self._cur_node_embedding(embeddings, td)
        state_embedding = self._state_embedding(embeddings, td)
        context_embedding = torch.cat([cur_node_embedding, state_embedding], -1)
        return self.project_context(context_embedding)


class FFSPContext(EnvContext):
    def __init__(self, embed_dim, stage_cnt=None):
        self.has_stage_emb = stage_cnt is not None
        step_context_dim = (1 + int(self.has_stage_emb)) * embed_dim
        super().__init__(embed_dim=embed_dim, step_context_dim=step_context_dim)
        if self.has_stage_emb:
            self.stage_emb = nn.Parameter(torch.rand(stage_cnt, embed_dim))

    def _cur_node_embedding(self, embeddings: TensorDict, td):
        cur_node_embedding = gather_by_index(
            embeddings["machine_embeddings"], td["stage_machine_idx"]
        )
        return cur_node_embedding

    def forward(self, embeddings, td):
        cur_node_embedding = self._cur_node_embedding(embeddings, td)
        if self.has_stage_emb:
            state_embedding = self._state_embedding(embeddings, td)
            context_embedding = torch.cat([cur_node_embedding, state_embedding], -1)
            return self.project_context(context_embedding)
        else:
            return self.project_context(cur_node_embedding)

    def _state_embedding(self, _, td):
        cur_stage_emb = self.stage_emb[td["stage_idx"]]
        return cur_stage_emb


class TSPContext(EnvContext):
    """Context embedding for the Traveling Salesman Problem (TSP).
    Project the following to the embedding space:
        - first node embedding
        - current node embedding
    """

    def __init__(self, embed_dim):
        super(TSPContext, self).__init__(embed_dim, 2 * embed_dim)
        self.W_placeholder = nn.Parameter(
            torch.Tensor(2 * self.embed_dim).uniform_(-1, 1)
        )

    def forward(self, embeddings, td):
        batch_size = embeddings.size(0)
        # By default, node_dim = -1 (we only have one node embedding per node)
        node_dim = (
            (-1,) if td["first_node"].dim() == 1 else (td["first_node"].size(-1), -1)
        )
        if td["i"][(0,) * td["i"].dim()].item() < 1:  # get first item fast
            if len(td.batch_size) < 2:
                context_embedding = self.W_placeholder[None, :].expand(
                    batch_size, self.W_placeholder.size(-1)
                )
            else:
                context_embedding = self.W_placeholder[None, None, :].expand(
                    batch_size, td.batch_size[1], self.W_placeholder.size(-1)
                )
        else:
            context_embedding = gather_by_index(
                embeddings,
                torch.stack([td["first_node"], td["current_node"]], -1).view(
                    batch_size, -1
                ),
            ).view(batch_size, *node_dim)
        return self.project_context(context_embedding)


class VRPContext(EnvContext):
    """Context embedding for the Capacitated Vehicle Routing Problem (CVRP).
    Project the following to the embedding space:
        - current node embedding
        - remaining capacity (vehicle_capacity - used_capacity)
    """

    def __init__(self, embed_dim):
        super(VRPContext, self).__init__(
            embed_dim=embed_dim, step_context_dim=embed_dim + 1
        )

    def _state_embedding(self, embeddings, td):
        state_embedding = td["vehicle_capacity"] - td["used_capacity"]
        return state_embedding


class VRPTWContext(VRPContext):
    """Context embedding for the Capacitated Vehicle Routing Problem (CVRP).
    Project the following to the embedding space:
        - current node embedding
        - remaining capacity (vehicle_capacity - used_capacity)
        - current time
    """

    def __init__(self, embed_dim):
        super(VRPContext, self).__init__(
            embed_dim=embed_dim, step_context_dim=embed_dim + 2
        )

    def _state_embedding(self, embeddings, td):
        capacity = super()._state_embedding(embeddings, td)
        current_time = td["current_time"]
        return torch.cat([capacity, current_time], -1)


class SVRPContext(EnvContext):
    """Context embedding for the Skill Vehicle Routing Problem (SVRP).
    Project the following to the embedding space:
        - current node embedding
        - current technician
    """

    def __init__(self, embed_dim):
        super(SVRPContext, self).__init__(embed_dim=embed_dim, step_context_dim=embed_dim)

    def forward(self, embeddings, td):
        cur_node_embedding = self._cur_node_embedding(embeddings, td).squeeze()
        return self.project_context(cur_node_embedding)


class PCTSPContext(EnvContext):
    """Context embedding for the Prize Collecting TSP (PCTSP).
    Project the following to the embedding space:
        - current node embedding
        - remaining prize (prize_required - cur_total_prize)
    """

    def __init__(self, embed_dim):
        super(PCTSPContext, self).__init__(embed_dim, embed_dim + 1)

    def _state_embedding(self, embeddings, td):
        state_embedding = torch.clamp(
            td["prize_required"] - td["cur_total_prize"], min=0
        )[..., None]
        return state_embedding


class OPContext(EnvContext):
    """Context embedding for the Orienteering Problem (OP).
    Project the following to the embedding space:
        - current node embedding
        - remaining distance (max_length - tour_length)
    """

    def __init__(self, embed_dim):
        super(OPContext, self).__init__(embed_dim, embed_dim + 1)

    def _state_embedding(self, embeddings, td):
        state_embedding = td["max_length"][..., 0] - td["tour_length"]
        return state_embedding[..., None]


class DPPContext(EnvContext):
    """Context embedding for the Decap Placement Problem (DPP), EDA (electronic design automation).
    Project the following to the embedding space:
        - current cell embedding
    """

    def __init__(self, embed_dim):
        super(DPPContext, self).__init__(embed_dim)

    def forward(self, embeddings, td):
        """Context cannot be defined by a single node embedding for DPP, hence 0.
        We modify the dynamic embedding instead to capture placed items
        """
        return embeddings.new_zeros(embeddings.size(0), self.embed_dim)


class PDPContext(EnvContext):
    """Context embedding for the Pickup and Delivery Problem (PDP).
    Project the following to the embedding space:
        - current node embedding
    """

    def __init__(self, embed_dim):
        super(PDPContext, self).__init__(embed_dim, embed_dim)

    def forward(self, embeddings, td):
        cur_node_embedding = self._cur_node_embedding(embeddings, td).squeeze()
        return self.project_context(cur_node_embedding)


class MTSPContext(EnvContext):
    """Context embedding for the Multiple Traveling Salesman Problem (mTSP).
    Project the following to the embedding space:
        - current node embedding
        - remaining_agents
        - current_length
        - max_subtour_length
        - distance_from_depot
    """

    def __init__(self, embed_dim, linear_bias=False):
        super(MTSPContext, self).__init__(embed_dim, 2 * embed_dim)
        proj_in_dim = (
            4  # remaining_agents, current_length, max_subtour_length, distance_from_depot
        )
        self.proj_dynamic_feats = nn.Linear(proj_in_dim, embed_dim, bias=linear_bias)

    def _cur_node_embedding(self, embeddings, td):
        cur_node_embedding = gather_by_index(embeddings, td["current_node"])
        return cur_node_embedding.squeeze()

    def _state_embedding(self, embeddings, td):
        dynamic_feats = torch.stack(
            [
                (td["num_agents"] - td["agent_idx"]).float(),
                td["current_length"],
                td["max_subtour_length"],
                self._distance_from_depot(td),
            ],
            dim=-1,
        )
        return self.proj_dynamic_feats(dynamic_feats)

    def _distance_from_depot(self, td):
        # Euclidean distance from the depot (loc[..., 0, :])
        cur_loc = gather_by_index(td["locs"], td["current_node"])
        return torch.norm(cur_loc - td["locs"][..., 0, :], dim=-1)


class SMTWTPContext(EnvContext):
    """Context embedding for the Single Machine Total Weighted Tardiness Problem (SMTWTP).
    Project the following to the embedding space:
        - current node embedding
        - current time
    """

    def __init__(self, embed_dim):
        super(SMTWTPContext, self).__init__(embed_dim, embed_dim + 1)

    def _cur_node_embedding(self, embeddings, td):
        cur_node_embedding = gather_by_index(embeddings, td["current_job"])
        return cur_node_embedding

    def _state_embedding(self, embeddings, td):
        state_embedding = td["current_time"]
        return state_embedding


class MDCPDPContext(EnvContext):
    """Context embedding for the MDCPDP.
    Project the following to the embedding space:
        - current node embedding
    """

    def __init__(self, embed_dim):
        super(MDCPDPContext, self).__init__(embed_dim, embed_dim * 2 + 5)

    def _state_embedding(self, embeddings, td):
        # get number of visited cities over total
        num_agents = td["capacity"].shape[-1]
        num_cities = td["locs"].shape[-2] - num_agents
        unvisited_number = td["available"][..., num_agents:].sum(-1)
        agent_capacity = td["capacity"].gather(-1, td["current_depot"])
        current_to_deliver = td["to_deliver"][..., num_agents + num_cities // 2 :]

        context_feats = torch.cat(
            [
                agent_capacity - td["current_carry"],  # current available capacity
                td["current_length"].gather(-1, td["current_depot"]),
                unvisited_number[..., None] / num_cities,
                current_to_deliver.sum(-1)[..., None],  # current to deliver number
                td["current_length"].max(-1)[0][..., None],  # max length
            ],
            -1,
        )
        return context_feats

    def _cur_agent_embedding(self, embeddings, td):
        """Get embedding of current agent"""
        cur_agent_embedding = gather_by_index(embeddings, td["current_depot"])
        return cur_agent_embedding

    def forward(self, embeddings, td):
        cur_node_embedding = self._cur_node_embedding(embeddings, td)
        cur_agent_embedding = self._cur_agent_embedding(embeddings, td)
        state_embedding = self._state_embedding(embeddings, td)
        context_embedding = torch.cat(
            [cur_node_embedding, cur_agent_embedding, state_embedding], -1
        )
        return self.project_context(context_embedding)


class SchedulingContext(nn.Module):
    def __init__(self, embed_dim: int, scaling_factor: int = 1000):
        super().__init__()
        self.scaling_factor = scaling_factor
        self.proj_busy = nn.Linear(1, embed_dim, bias=False)

    def forward(self, h, td):
        busy_for = (td["busy_until"] - td["time"].unsqueeze(1)) / self.scaling_factor
        busy_proj = self.proj_busy(busy_for.unsqueeze(-1))
        # (b m e)
        return h + busy_proj


class MTVRPContext(VRPContext):
    """Context embedding for Multi-Task VRPEnv.
    Project the following to the embedding space:
        - current node embedding
        - remaining_linehaul_capacity (vehicle_capacity - used_capacity_linehaul)
        - remaining_backhaul_capacity (vehicle_capacity - used_capacity_backhaul)
        - current time
        - current_route_length
        - open route indicator
    """

    def __init__(self, embed_dim):
        super(VRPContext, self).__init__(
            embed_dim=embed_dim, step_context_dim=embed_dim + 5
        )

    def _state_embedding(self, embeddings, td):
        remaining_linehaul_capacity = (
            td["vehicle_capacity"] - td["used_capacity_linehaul"]
        )
        remaining_backhaul_capacity = (
            td["vehicle_capacity"] - td["used_capacity_backhaul"]
        )
        current_time = td["current_time"]
        current_route_length = td["current_route_length"]
        open_route = td["open_route"]
        return torch.cat(
            [
                remaining_linehaul_capacity,
                remaining_backhaul_capacity,
                current_time,
                current_route_length,
                open_route,
            ],
            -1,
        )


class FLPContext(EnvContext):
    """Context embedding for the Facility Location Problem (FLP)."""

    def __init__(self, embed_dim: int):
        super(FLPContext, self).__init__(embed_dim=embed_dim)
        self.embed_dim = embed_dim
        self.project_context = nn.Linear(embed_dim, embed_dim, bias=True)

    def forward(self, embeddings, td):
        cur_dist = td["distances"].unsqueeze(-2)  # (batch_size, 1, n_points)
        dist_improve = cur_dist - td["orig_distances"]  # (batch_size, n_points, n_points)
        dist_improve = torch.clamp(dist_improve, min=0).sum(-1)  # (batch_size, n_points)

        # softmax
        loc_best_soft = torch.softmax(dist_improve, dim=-1)  # (batch_size, n_points)
        context_embedding = (embeddings * loc_best_soft[..., None]).sum(-2)
        return self.project_context(context_embedding)


class MCPContext(EnvContext):
    """Context embedding for the Maximum Coverage Problem (MCP)."""

    def __init__(self, embed_dim: int):
        super(MCPContext, self).__init__(embed_dim=embed_dim)
        self.embed_dim = embed_dim
        self.project_context = nn.Linear(embed_dim, embed_dim, bias=True)

    def forward(self, embeddings, td):
        membership_weighted = batched_scatter_sum(
            td["weights"].unsqueeze(-1), td["membership"].long()
        )
        membership_weighted.squeeze_(-1)
        # membership_weighted: [batch_size, n_sets]

        # softmax; higher weights for better sets
        membership_weighted = torch.softmax(
            membership_weighted, dim=-1
        )  # (batch_size, n_sets)
        context_embedding = (membership_weighted.unsqueeze(-1) * embeddings).sum(1)
        return self.project_context(context_embedding)



from typing import Any, Callable

import torch.nn as nn

from rl4co.data.transforms import StateAugmentation
from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl.reinforce.reinforce import REINFORCE
from rl4co.models.zoo.symnco.losses import (
    invariance_loss,
    problem_symmetricity_loss,
    solution_symmetricity_loss,
)
from rl4co.models.zoo.symnco.policy import SymNCOPolicy
from rl4co.utils.ops import gather_by_index, get_num_starts, unbatchify
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class SymNCO(REINFORCE):
    """SymNCO Model based on REINFORCE with shared baselines.
    Based on Kim et al. (2022) https://arxiv.org/abs/2205.13209.

    Args:
        env: TorchRL environment to use for the algorithm
        policy: Policy to use for the algorithm
        policy_kwargs: Keyword arguments for policy
        num_augment: Number of augmentations
        augment_fn: Function to use for augmentation, defaulting to dihedral_8_augmentation
        feats: List of features to augment
        alpha: weight for invariance loss
        beta: weight for solution symmetricity loss
        num_starts: Number of starts for multi-start. If None, use the number of available actions
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: nn.Module | SymNCOPolicy = None,
        policy_kwargs: dict = {},
        baseline: str = "symnco",
        num_augment: int = 4,
        augment_fn: str | Callable = "symmetric",
        feats: list = None,
        alpha: float = 0.2,
        beta: float = 1,
        num_starts: int = 0,
        **kwargs,
    ):
        self.save_hyperparameters(logger=False)

        if policy is None:
            policy = SymNCOPolicy(env_name=env.name, **policy_kwargs)

        assert baseline == "symnco", "SymNCO only supports custom-symnco baseline"
        baseline = "no"  # Pass no baseline to superclass since there are multiple custom baselines

        # Pass no baseline to superclass since there are multiple custom baselines
        super().__init__(env, policy, baseline, **kwargs)

        self.num_starts = num_starts
        self.num_augment = num_augment
        self.augment = StateAugmentation(
            num_augment=self.num_augment, augment_fn=augment_fn, feats=feats
        )
        self.alpha = alpha  # weight for invariance loss
        self.beta = beta  # weight for solution symmetricity loss

        # Add `_multistart` to decode type for train, val and test in policy if num_starts > 1
        if self.num_starts > 1:
            for phase in ["train", "val", "test"]:
                self.set_decode_type_multistart(phase)

    def shared_step(
        self, batch: Any, batch_idx: int, phase: str, dataloader_idx: int = None
    ):
        td = self.env.reset(batch)
        n_aug, n_start = self.num_augment, self.num_starts
        n_start = get_num_starts(td, self.env.name) if n_start is None else n_start

        # Symmetric augmentation
        if n_aug > 1:
            td = self.augment(td)

        # Evaluate policy
        out = self.policy(td, self.env, phase=phase, num_starts=n_start)

        # Unbatchify reward to [batch_size, n_aug, n_start].
        unbatch_dims = (n_aug, n_start)
        reward = unbatchify(out["reward"], unbatch_dims)

        # Main training loss
        if phase == "train":
            # [batch_size, n_start, n_aug]
            ll = unbatchify(out["log_likelihood"], unbatch_dims)

            # Calculate losses: problem symmetricity, solution symmetricity, invariance
            loss_ps = problem_symmetricity_loss(reward, ll) if n_start > 1 else 0
            loss_ss = solution_symmetricity_loss(reward, ll) if n_aug > 1 else 0
            loss_inv = invariance_loss(out["proj_embeddings"], n_aug) if n_aug > 1 else 0
            loss = loss_ps + self.beta * loss_ss + self.alpha * loss_inv
            out.update(
                {
                    "loss": loss,
                    "loss_ss": loss_ss,
                    "loss_ps": loss_ps,
                    "loss_inv": loss_inv,
                }
            )

        # Log only during validation and test
        else:
            if n_start > 1:
                # max multi-start reward
                max_reward, max_idxs = reward.max(dim=1)
                out.update({"max_reward": max_reward})

                # Reshape batch to [batch, n_start, n_aug]
                if out.get("actions", None) is not None:
                    actions = unbatchify(out["actions"], unbatch_dims)
                    out.update(
                        {"best_multistart_actions": gather_by_index(actions, max_idxs)}
                    )
                    out["actions"] = actions

            # Get augmentation score only during inference
            if n_aug > 1:
                # If multistart is enabled, we use the best multistart rewards
                reward_ = max_reward if n_start > 1 else reward
                max_aug_reward, max_idxs = reward_.max(dim=1)
                out.update({"max_aug_reward": max_aug_reward})
                if out.get("best_multistart_actions", None) is not None:
                    out.update(
                        {
                            "best_aug_actions": gather_by_index(
                                out["best_multistart_actions"], max_idxs
                            )
                        }
                    )

        metrics = self.log_metrics(out, phase, dataloader_idx=dataloader_idx)
        return {"loss": out.get("loss", None), **metrics}

    @classmethod
    def load_from_checkpoint(
        cls,
        checkpoint_path,
        map_location=None,
        hparams_file=None,
        strict=False,
        load_baseline=True,
        **kwargs,
    ):
        if kwargs.pop("baseline", "symnco") != "symnco":
            log.warning(
                "SymNCO only supports custom-symnco baseline. Setting to 'symnco'."
            )
        kwargs["baseline"] = "symnco"
        return super().load_from_checkpoint(
            checkpoint_path,
            map_location,
            hparams_file,
            strict,
            load_baseline,
            **kwargs,
        )



from typing import Optional

import torch.nn as nn

from rl4co.models.common.constructive.nonautoregressive import (
    NonAutoregressiveDecoder,
    NonAutoregressiveEncoder,
    NonAutoregressivePolicy,
)
from rl4co.utils.pylogger import get_pylogger

from .encoder import NARGNNEncoder

log = get_pylogger(__name__)


class NARGNNPolicy(NonAutoregressivePolicy):
    """
    Base Non-autoregressive policy for NCO construction methods.
    This creates a heatmap of NxN for N nodes (i.e., heuristic) that models the probability to go from one node to another for all nodes.

    The policy performs the following steps:
        1. Encode the environment initial state into node embeddings
        2. Decode (non-autoregressively) to construct the solution to the NCO problem

    Warning:
        The effectiveness of the non-autoregressive approach can vary significantly across different problem types and configurations.
        It may require careful tuning of the model architecture and decoding strategy to achieve competitive results.

    Args:
        encoder: Encoder module. Can be passed by sub-classes
        decoder: Decoder module. Note that this moule defaults to the non-autoregressive decoder
        embed_dim: Dimension of the embeddings
        env_name: Name of the environment used to initialize embeddings
        init_embedding: Model to use for the initial embedding. If None, use the default embedding for the environment
        edge_embedding: Model to use for the edge embedding. If None, use the default embedding for the environment
        graph_network: Model to use for the graph network. If None, use the default embedding for the environment
        heatmap_generator: Model to use for the heatmap generator. If None, use the default embedding for the environment
        num_layers_heatmap_generator: Number of layers in the heatmap generator
        num_layers_graph_encoder: Number of layers in the graph encoder
        act_fn: Activation function to use in the encoder
        agg_fn: Aggregation function to use in the encoder
        linear_bias: Whether to use bias in the encoder
        train_decode_type: Type of decoding during training
        val_decode_type: Type of decoding during validation
        test_decode_type: Type of decoding during testing
        **constructive_policy_kw: Unused keyword arguments
    """

    def __init__(
        self,
        encoder: Optional[NonAutoregressiveEncoder] = None,
        decoder: Optional[NonAutoregressiveDecoder] = None,
        embed_dim: int = 64,
        env_name: str = "tsp",
        init_embedding: Optional[nn.Module] = None,
        edge_embedding: Optional[nn.Module] = None,
        graph_network: Optional[nn.Module] = None,
        heatmap_generator: Optional[nn.Module] = None,
        num_layers_heatmap_generator: int = 5,
        num_layers_graph_encoder: int = 15,
        act_fn="silu",
        agg_fn="mean",
        linear_bias: bool = True,
        train_decode_type: str = "multistart_sampling",
        val_decode_type: str = "multistart_greedy",
        test_decode_type: str = "multistart_greedy",
        **constructive_policy_kw,
    ):
        if len(constructive_policy_kw) > 0:
            log.warn(f"Unused kwargs: {constructive_policy_kw}")

        if encoder is None:
            encoder = NARGNNEncoder(
                embed_dim=embed_dim,
                env_name=env_name,
                init_embedding=init_embedding,
                edge_embedding=edge_embedding,
                graph_network=graph_network,
                heatmap_generator=heatmap_generator,
                num_layers_heatmap_generator=num_layers_heatmap_generator,
                num_layers_graph_encoder=num_layers_graph_encoder,
                act_fn=act_fn,
                agg_fn=agg_fn,
                linear_bias=linear_bias,
            )

        # The decoder generates logits given the current td and heatmap
        if decoder is None:
            decoder = NonAutoregressiveDecoder()
        else:
            # check if the decoder has trainable parameters
            if any(p.requires_grad for p in decoder.parameters()):
                log.error(
                    "The decoder contains trainable parameters. This should not happen in a non-autoregressive policy."
                )

        # Pass to constructive policy
        super(NARGNNPolicy, self).__init__(
            encoder=encoder,
            decoder=decoder,
            env_name=env_name,
            train_decode_type=train_decode_type,
            val_decode_type=val_decode_type,
            test_decode_type=test_decode_type,
            **constructive_policy_kw,
        )



from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.data.utils import load_npz_to_tensordict
from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.ops import gather_by_index, get_distance
from rl4co.utils.pylogger import get_pylogger

from .generator import MTVRPGenerator

log = get_pylogger(__name__)


class MTVRPEnv(RL4COEnvBase):
    r"""MTVRPEnv is a Multi-Task VRP environment which can take any combination of the following constraints:

    Features:

    - *Capacity (C)*
        - Each vehicle has a maximum capacity $Q$, restricting the total load that can be in the vehicle at any point of the route.
        - The route must be planned such that the sum of demands and pickups for all customers visited does not exceed this capacity.
    - *Time Windows (TW)*
        - Every node $i$ has an associated time window $[e_i, l_i]$ during which service must commence.
        - Additionally, each node has a service time $s_i$. Vehicles must reach node $i$ within its time window; early arrivals must wait at the node location until time $e_i$.
    - *Open Routes (O)*
        - Vehicles are not required to return to the depot after serving all customers.
        - Note that this does not need to be counted as a constraint since it can be modelled by setting zero costs on arcs returning to the depot $c_{i0} = 0$ from any customer $i \in C$, and not counting the return arc as part of the route.
    - *Backhauls (B)*
        - Backhauls generalize demand to also account for return shipments. Customers are either linehaul or backhaul customers.
        - Linehaul customers require delivery of a demand $q_i > 0$ that needs to be transported from the depot to the customer, whereas backhaul customers need a pickup of an amount $p_i > 0$ that is transported from the client back to the depot.
        - It is possible for vehicles to serve a combination of linehaul and backhaul customers in a single route, but then any linehaul customers must precede the backhaul customers in the route.
    - *Duration Limits (L)*
        - Imposes a limit on the total travel duration (or length) of each route, ensuring a balanced workload across vehicles.

    The environment covers the following 16 variants depending on the data generation:

    | VRP Variant | Capacity (C) | Open Route (O) | Backhaul (B) | Duration Limit (L) | Time Window (TW) |
    | :---------- | :----------: | :------------: | :----------: | :----------------: | :--------------: |
    | CVRP        |      ✔       |                |              |                    |                  |
    | OVRP        |      ✔       |       ✔        |              |                    |                  |
    | VRPB        |      ✔       |                |      ✔       |                    |                  |
    | VRPL        |      ✔       |                |              |         ✔          |                  |
    | VRPTW       |      ✔       |                |              |                    |        ✔         |
    | OVRPTW      |      ✔       |       ✔        |              |                    |        ✔         |
    | OVRPB       |      ✔       |       ✔        |      ✔       |                    |                  |
    | OVRPL       |      ✔       |       ✔        |              |         ✔          |                  |
    | VRPBL       |      ✔       |                |      ✔       |         ✔          |                  |
    | VRPBTW      |      ✔       |                |      ✔       |                    |        ✔         |
    | VRPLTW      |      ✔       |                |              |         ✔          |        ✔         |
    | OVRPBL      |      ✔       |       ✔        |      ✔       |         ✔          |                  |
    | OVRPBTW     |      ✔       |       ✔        |      ✔       |                    |        ✔         |
    | OVRPLTW     |      ✔       |       ✔        |              |         ✔          |        ✔         |
    | VRPBLTW     |      ✔       |                |      ✔       |         ✔          |        ✔         |
    | OVRPBLTW    |      ✔       |       ✔        |      ✔       |         ✔          |        ✔         |

    You may also check out the following papers as reference:
    - ["Multi-Task Learning for Routing Problem with Cross-Problem Zero-Shot Generalization" (Liu et al, 2024)](https://arxiv.org/abs/2402.16891)
    - ["MVMoE: Multi-Task Vehicle Routing Solver with Mixture-of-Experts" (Zhou et al, 2024)](https://arxiv.org/abs/2405.01029)
    - ["RouteFinder: Towards Foundation Models for Vehicle Routing Problems" (Berto et al, 2024)](https://arxiv.org/abs/2406.15007)

    Tip:
        Have a look at https://pyvrp.org/ for more information about VRP and its variants and their solutions. Kudos to their help and great job!

    Args:
        generator: Generator for the environment, see :class:`MTVRPGenerator`.
        generator_params: Parameters for the generator.
    """

    name = "mtvrp"

    def __init__(
        self,
        generator: MTVRPGenerator = None,
        generator_params: dict = {},
        check_solution: bool = False,
        **kwargs,
    ):
        if check_solution:
            log.warning(
                "Solution checking is enabled. This may slow down the environment."
                " We recommend disabling this for training by passing `check_solution=False`."
            )

        super().__init__(check_solution=check_solution, **kwargs)

        if generator is None:
            generator = MTVRPGenerator(**generator_params)
        self.generator = generator
        self._make_spec(self.generator)

    def _step(self, td: TensorDict) -> TensorDict:
        # Get locations and distance
        prev_node, curr_node = td["current_node"], td["action"]
        prev_loc = gather_by_index(td["locs"], prev_node)
        curr_loc = gather_by_index(td["locs"], curr_node)
        distance = get_distance(prev_loc, curr_loc)[..., None]

        # Update current time
        service_time = gather_by_index(
            src=td["service_time"], idx=curr_node, dim=1, squeeze=False
        )
        start_times = gather_by_index(
            src=td["time_windows"], idx=curr_node, dim=1, squeeze=False
        )[..., 0]
        # we cannot start before we arrive and we should start at least at start times
        curr_time = (curr_node[:, None] != 0) * (
            torch.max(td["current_time"] + distance / td["speed"], start_times)
            + service_time
        )

        # Update current route length (reset at depot)
        curr_route_length = (curr_node[:, None] != 0) * (
            td["current_route_length"] + distance
        )

        # Linehaul (delivery) demands
        selected_demand_linehaul = gather_by_index(
            td["demand_linehaul"], curr_node, dim=1, squeeze=False
        )
        selected_demand_backhaul = gather_by_index(
            td["demand_backhaul"], curr_node, dim=1, squeeze=False
        )

        # Backhaul (pickup) demands
        # vehicles are empty once we get to the backhauls
        used_capacity_linehaul = (curr_node[:, None] != 0) * (
            td["used_capacity_linehaul"] + selected_demand_linehaul
        )
        used_capacity_backhaul = (curr_node[:, None] != 0) * (
            td["used_capacity_backhaul"] + selected_demand_backhaul
        )

        # Done when all customers are visited
        visited = td["visited"].scatter(-1, curr_node[..., None], True)
        done = visited.sum(-1) == visited.size(-1)
        reward = torch.zeros_like(
            done
        ).float()  # we use the `get_reward` method to compute the reward

        td.update(
            {
                "current_node": curr_node,
                "current_route_length": curr_route_length,
                "current_time": curr_time,
                "done": done,
                "reward": reward,
                "used_capacity_linehaul": used_capacity_linehaul,
                "used_capacity_backhaul": used_capacity_backhaul,
                "visited": visited,
            }
        )
        td.set("action_mask", self.get_action_mask(td))
        return td

    def _reset(
        self,
        td: Optional[TensorDict] = None,
        batch_size: Optional[list] = None,
    ) -> TensorDict:
        device = td.device

        # Create reset TensorDict
        td_reset = TensorDict(
            {
                "locs": td["locs"],
                "demand_backhaul": td["demand_backhaul"],
                "demand_linehaul": td["demand_linehaul"],
                "distance_limit": td["distance_limit"],
                "service_time": td["service_time"],
                "open_route": td["open_route"],
                "time_windows": td["time_windows"],
                "vehicle_capacity": td["vehicle_capacity"],
                "capacity_original": td["capacity_original"],
                "speed": td["speed"],
                "current_node": torch.zeros(
                    (*batch_size,), dtype=torch.long, device=device
                ),
                "current_route_length": torch.zeros(
                    (*batch_size, 1), dtype=torch.float32, device=device
                ),  # for distance limits
                "current_time": torch.zeros(
                    (*batch_size, 1), dtype=torch.float32, device=device
                ),  # for time windows
                "used_capacity_backhaul": torch.zeros(
                    (*batch_size, 1), device=device
                ),  # for capacity constraints in backhaul
                "used_capacity_linehaul": torch.zeros(
                    (*batch_size, 1), device=device
                ),  # for capacity constraints in linehaul
                "visited": torch.zeros(
                    (*batch_size, td["locs"].shape[-2]),
                    dtype=torch.bool,
                    device=device,
                ),
            },
            batch_size=batch_size,
            device=device,
        )
        td_reset.set("action_mask", self.get_action_mask(td_reset))
        return td_reset

    @staticmethod
    def get_action_mask(td: TensorDict) -> torch.Tensor:
        curr_node = td["current_node"]  # note that this was just updated!
        locs = td["locs"]
        d_ij = get_distance(
            gather_by_index(locs, curr_node)[..., None, :], locs
        )  # i (current) -> j (next)
        d_j0 = get_distance(locs, locs[..., 0:1, :])  # j (next) -> 0 (depot)

        # Time constraint (TW):
        early_tw, late_tw = (
            td["time_windows"][..., 0],
            td["time_windows"][..., 1],
        )
        arrival_time = td["current_time"] + (d_ij / td["speed"])
        # can reach in time -> only need to *start* in time
        can_reach_customer = arrival_time < late_tw
        # we must ensure that we can return to depot in time *if* route is closed
        # i.e. start time + service time + time back to depot < late_tw
        can_reach_depot = (
            torch.max(arrival_time, early_tw) + td["service_time"] + (d_j0 / td["speed"])
        ) * ~td["open_route"] < late_tw[..., 0:1]

        # Distance limit (L): do not add distance to depot if open route (O)
        exceeds_dist_limit = (
            td["current_route_length"] + d_ij + (d_j0 * ~td["open_route"])
            > td["distance_limit"]
        )

        # Linehaul demand / delivery (C) and backhaul demand / pickup (B)
        # All linehauls are visited before backhauls
        linehauls_missing = ((td["demand_linehaul"] * ~td["visited"]).sum(-1) > 0)[
            ..., None
        ]
        is_carrying_backhaul = (
            gather_by_index(
                src=td["demand_backhaul"],
                idx=curr_node,
                dim=1,
                squeeze=False,
            )
            > 0
        )
        exceeds_cap_linehaul = (
            td["demand_linehaul"] + td["used_capacity_linehaul"] > td["vehicle_capacity"]
        )
        exceeds_cap_backhaul = (
            td["demand_backhaul"] + td["used_capacity_backhaul"] > td["vehicle_capacity"]
        )

        meets_demand_constraint = (
            linehauls_missing
            & ~exceeds_cap_linehaul
            & ~is_carrying_backhaul
            & (td["demand_linehaul"] > 0)
        ) | (~exceeds_cap_backhaul & (td["demand_backhaul"] > 0))

        # Condense constraints
        can_visit = (
            can_reach_customer
            & can_reach_depot
            & meets_demand_constraint
            & ~exceeds_dist_limit
            & ~td["visited"]
        )

        # Mask depot: don't visit depot if coming from there and there are still customer nodes I can visit
        can_visit[:, 0] = ~((curr_node == 0) & (can_visit[:, 1:].sum(-1) > 0))
        return can_visit

    def _get_reward(self, td: TensorDict, actions: TensorDict) -> TensorDict:
        # Append depot to actions and get sequence of locations
        go_from = torch.cat((torch.zeros_like(actions[:, :1]), actions), dim=1)
        go_to = torch.roll(go_from, -1, dims=1)  # [b, seq_len]
        loc_from = gather_by_index(td["locs"], go_from)
        loc_to = gather_by_index(td["locs"], go_to)

        # Get tour length. If route is open and goes to depot, don't count the distance
        distances = get_distance(loc_from, loc_to)  # [b, seq_len]
        tour_length = (distances * ~((go_to == 0) & td["open_route"])).sum(-1)  # [b]
        return -tour_length  # reward is negative cost

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor):
        batch_size, n_loc = td["demand_linehaul"].size()
        locs = td["locs"]
        n_loc -= 1  # exclude depot
        sorted_pi = actions.data.sort(1)[0]

        # all customer nodes visited exactly once
        assert (
            torch.arange(1, n_loc + 1, out=sorted_pi.data.new())
            .view(1, -1)
            .expand(batch_size, n_loc)
            == sorted_pi[:, -n_loc:]
        ).all() and (sorted_pi[:, :-n_loc] == 0).all(), "Invalid tour"

        # Distance limits (L)
        assert (td["distance_limit"] >= 0).all(), "Distance limits must be non-negative."

        # Time windows (TW)
        d_j0 = get_distance(locs, locs[..., 0:1, :])  # j (next) -> 0 (depot)
        assert torch.all(td["time_windows"] >= 0.0), "Time windows must be non-negative."
        assert torch.all(td["service_time"] >= 0.0), "Service time must be non-negative."
        assert torch.all(
            td["time_windows"][..., 0] < td["time_windows"][..., 1]
        ), "there are unfeasible time windows"
        assert torch.all(
            td["time_windows"][..., :, 0] + d_j0 + td["service_time"]
            <= td["time_windows"][..., 0, 1, None]
        ), "vehicle cannot perform service and get back to depot in time."
        # check individual time windows
        curr_time = torch.zeros(batch_size, dtype=torch.float32, device=td.device)
        curr_node = torch.zeros(batch_size, dtype=torch.int64, device=td.device)
        curr_length = torch.zeros(batch_size, dtype=torch.float32, device=td.device)
        for ii in range(actions.size(1)):
            next_node = actions[:, ii]
            curr_loc = gather_by_index(td["locs"], curr_node)
            next_loc = gather_by_index(td["locs"], next_node)
            dist = get_distance(curr_loc, next_loc)

            # distance limit (L)
            curr_length = curr_length + dist * ~(
                td["open_route"].squeeze(-1) & (next_node == 0)
            )  # do not count back to depot for open route
            assert torch.all(
                curr_length <= td["distance_limit"].squeeze(-1)
            ), "Route exceeds distance limit"
            curr_length[next_node == 0] = 0.0  # reset length for depot

            curr_time = torch.max(
                curr_time + dist, gather_by_index(td["time_windows"], next_node)[..., 0]
            )
            assert torch.all(
                curr_time <= gather_by_index(td["time_windows"], next_node)[..., 1]
            ), "vehicle cannot start service before deadline"
            curr_time = curr_time + gather_by_index(td["service_time"], next_node)
            curr_node = next_node
            curr_time[curr_node == 0] = 0.0  # reset time for depot

        # Demand constraints (C) and (B)
        # linehauls are the same as backhauls but with a different feature
        def _check_c1(feature="demand_linehaul"):
            demand = td[feature].gather(dim=1, index=actions)
            used_cap = torch.zeros_like(td[feature][:, 0])
            for ii in range(actions.size(1)):
                # reset at depot
                used_cap = used_cap * (actions[:, ii] != 0)
                used_cap += demand[:, ii]
                assert (
                    used_cap <= td["vehicle_capacity"]
                ).all(), "Used more than capacity for {}: {}".format(feature, used_cap)

        _check_c1("demand_linehaul")
        _check_c1("demand_backhaul")

    def load_data(self, fpath, batch_size=[], scale=False):
        """Dataset loading from file
        Normalize demand by capacity to be in [0, 1]
        """
        td_load = load_npz_to_tensordict(fpath)
        if scale:
            td_load.set(
                "demand_linehaul",
                td_load["demand_linehaul"] / td_load["capacity_original"],
            )
            td_load.set(
                "demand_backhaul",
                td_load["demand_backhaul"] / td_load["capacity_original"],
            )
        return td_load

    @staticmethod
    def render(*args, **kwargs):
        """Simple wrapper for render function"""
        from .render import render

        return render(*args, **kwargs)

    def select_start_nodes(self, td, num_starts):
        """Select available start nodes for the environment (e.g. for POMO-based training)"""
        num_loc = td["locs"].shape[-2] - 1
        selected = (
            torch.arange(num_starts, device=td.device).repeat_interleave(td.shape[0])
            % num_loc
            + 1
        )
        return selected

    @staticmethod
    def solve(
        instances: TensorDict,
        max_runtime: float,
        num_procs: int = 1,
        solver: str = "pyvrp",
        **kwargs,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Classical solver for the environment. This is a wrapper for the baselines solver.
        Available solvers are: `pyvrp`, `ortools`, `lkh`. Returns the actions and costs.
        """
        from .baselines.solve import solve

        return solve(instances, max_runtime, num_procs, solver, **kwargs)

    def _make_spec(self, td_params: TensorDict):
        # TODO: include extra vars (but we don't really need them for now)
        """Make the observation and action specs from the parameters."""
        self.observation_spec = Composite(
            locs=Bounded(
                low=self.generator.min_loc,
                high=self.generator.max_loc,
                shape=(self.generator.num_loc + 1, 2),
                dtype=torch.float32,
                device=self.device,
            ),
            current_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
                device=self.device,
            ),
            demand_linehaul=Bounded(
                low=-self.generator.capacity,
                high=self.generator.max_demand,
                shape=(self.generator.num_loc, 1),  # demand is only for customers
                dtype=torch.float32,
                device=self.device,
            ),
            demand_backhaul=Bounded(
                low=-self.generator.capacity,
                high=self.generator.max_demand,
                shape=(self.generator.num_loc, 1),  # demand is only for customers
                dtype=torch.float32,
                device=self.device,
            ),
            action_mask=Unbounded(
                shape=(self.generator.num_loc + 1, 1),
                dtype=torch.bool,
                device=self.device,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            low=0,
            high=self.generator.num_loc + 1,
            shape=(1,),
            dtype=torch.int64,
            device=self.device,
        )
        self.reward_spec = Unbounded(shape=(1,), dtype=torch.float32, device=self.device)
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool, device=self.device)

    @staticmethod
    def check_variants(td):
        """Check if the problem has the variants"""
        has_open = td["open_route"].squeeze(-1)
        has_tw = (td["time_windows"][:, :, 1] != float("inf")).any(-1)
        has_limit = (td["distance_limit"] != float("inf")).squeeze(-1)
        has_backhaul = (td["demand_backhaul"] != 0).any(-1)
        return has_open, has_tw, has_limit, has_backhaul

    @staticmethod
    def get_variant_names(td):
        (
            has_open,
            has_time_window,
            has_duration_limit,
            has_backhaul,
        ) = MTVRPEnv.check_variants(td)
        instance_names = []
        for o, b, l_, tw in zip(
            has_open, has_backhaul, has_duration_limit, has_time_window
        ):
            if not o and not b and not l_ and not tw:
                instance_name = "CVRP"
            else:
                instance_name = "VRP"
                if o:
                    instance_name = "O" + instance_name
                if b:
                    instance_name += "B"
                if l_:
                    instance_name += "L"
                if tw:
                    instance_name += "TW"
            instance_names.append(instance_name)
        return instance_names

    def print_presets(self):
        self.generator.print_presets()



from typing import Optional

import torch
import torch.nn as nn

from torch.distributions import Categorical

from rl4co.models.common.constructive.autoregressive import (
    AutoregressiveDecoder,
    AutoregressiveEncoder,
    AutoregressivePolicy,
)
from rl4co.models.common.constructive.base import NoEncoder
from rl4co.models.nn.env_embeddings.init import FJSPMatNetInitEmbedding
from rl4co.models.nn.graph.hgnn import HetGNNEncoder
from rl4co.models.nn.mlp import MLP
from rl4co.models.zoo.matnet.matnet_w_sa import Encoder
from rl4co.utils.decoding import DecodingStrategy, process_logits
from rl4co.utils.ops import gather_by_index
from rl4co.utils.pylogger import get_pylogger

from .decoder import L2DAttnActor, L2DDecoder
from .encoder import GCN4JSSP

log = get_pylogger(__name__)


class L2DPolicy(AutoregressivePolicy):
    def __init__(
        self,
        encoder: Optional[AutoregressiveEncoder] = None,
        decoder: Optional[AutoregressiveDecoder] = None,
        embed_dim: int = 64,
        num_encoder_layers: int = 2,
        env_name: str = "fjsp",
        het_emb: bool = True,
        scaling_factor: int = 1000,
        normalization: str = "batch",
        init_embedding: Optional[nn.Module] = None,
        stepwise_encoding: bool = False,
        tanh_clipping: float = 10,
        train_decode_type: str = "sampling",
        val_decode_type: str = "greedy",
        test_decode_type: str = "multistart_sampling",
        **constructive_policy_kw,
    ):
        if len(constructive_policy_kw) > 0:
            log.warn(f"Unused kwargs: {constructive_policy_kw}")

        if encoder is None:
            if stepwise_encoding:
                encoder = NoEncoder()
            elif env_name == "fjsp" or (env_name == "jssp" and het_emb):
                encoder = HetGNNEncoder(
                    env_name=env_name,
                    embed_dim=embed_dim,
                    num_layers=num_encoder_layers,
                    normalization="batch",
                    init_embedding=init_embedding,
                    scaling_factor=scaling_factor,
                )
            else:
                encoder = GCN4JSSP(
                    embed_dim,
                    num_encoder_layers,
                    init_embedding=init_embedding,
                    scaling_factor=scaling_factor,
                )

        # The decoder generates logits given the current td and heatmap
        if decoder is None:
            decoder = L2DDecoder(
                env_name=env_name,
                embed_dim=embed_dim,
                actor_hidden_dim=embed_dim,
                num_encoder_layers=num_encoder_layers,
                init_embedding=init_embedding,
                het_emb=het_emb,
                stepwise=stepwise_encoding,
                scaling_factor=scaling_factor,
                normalization=normalization,
            )

        # Pass to constructive policy
        super(L2DPolicy, self).__init__(
            encoder=encoder,
            decoder=decoder,
            env_name=env_name,
            tanh_clipping=tanh_clipping,
            train_decode_type=train_decode_type,
            val_decode_type=val_decode_type,
            test_decode_type=test_decode_type,
            **constructive_policy_kw,
        )


class L2DAttnPolicy(AutoregressivePolicy):
    def __init__(
        self,
        encoder: Optional[AutoregressiveEncoder] = None,
        decoder: Optional[AutoregressiveDecoder] = None,
        embed_dim: int = 256,
        num_heads: int = 8,
        num_encoder_layers: int = 4,
        scaling_factor: int = 1000,
        normalization: str = "batch",
        env_name: str = "fjsp",
        init_embedding: Optional[nn.Module] = None,
        tanh_clipping: float = 10,
        train_decode_type: str = "sampling",
        val_decode_type: str = "greedy",
        test_decode_type: str = "multistart_sampling",
        **constructive_policy_kw,
    ):
        if len(constructive_policy_kw) > 0:
            log.warn(f"Unused kwargs: {constructive_policy_kw}")

        if encoder is None:
            if init_embedding is None:
                init_embedding = FJSPMatNetInitEmbedding(
                    embed_dim, scaling_factor=scaling_factor
                )

            encoder = Encoder(
                embed_dim=embed_dim,
                num_heads=num_heads,
                num_layers=num_encoder_layers,
                normalization=normalization,
                feedforward_hidden=embed_dim * 2,
                init_embedding=init_embedding,
            )

        # The decoder generates logits given the current td and heatmap
        if decoder is None:
            decoder = L2DAttnActor(
                env_name=env_name,
                embed_dim=embed_dim,
                num_heads=num_heads,
                scaling_factor=scaling_factor,
                stepwise=False,
            )

        # Pass to constructive policy
        super(L2DAttnPolicy, self).__init__(
            encoder=encoder,
            decoder=decoder,
            env_name=env_name,
            tanh_clipping=tanh_clipping,
            train_decode_type=train_decode_type,
            val_decode_type=val_decode_type,
            test_decode_type=test_decode_type,
            **constructive_policy_kw,
        )


class L2DPolicy4PPO(L2DPolicy):
    def __init__(
        self,
        encoder=None,
        decoder=None,
        critic=None,
        embed_dim: int = 64,
        num_encoder_layers: int = 2,
        env_name: str = "fjsp",
        het_emb: bool = True,
        scaling_factor: int = 1000,
        init_embedding=None,
        tanh_clipping: float = 10,
        train_decode_type: str = "sampling",
        val_decode_type: str = "greedy",
        test_decode_type: str = "multistart_sampling",
        **constructive_policy_kw,
    ):
        if init_embedding is None:
            pass  # TODO PPO specific init emb?

        super().__init__(
            encoder=encoder,
            decoder=decoder,
            embed_dim=embed_dim,
            num_encoder_layers=num_encoder_layers,
            env_name=env_name,
            het_emb=het_emb,
            scaling_factor=scaling_factor,
            init_embedding=init_embedding,
            stepwise_encoding=True,
            tanh_clipping=tanh_clipping,
            train_decode_type=train_decode_type,
            val_decode_type=val_decode_type,
            test_decode_type=test_decode_type,
            **constructive_policy_kw,
        )

        if critic is None:
            if env_name == "fjsp" or het_emb:
                input_dim = 2 * embed_dim
            else:
                input_dim = embed_dim
            critic = MLP(input_dim, 1, num_neurons=[embed_dim] * 2)

        self.critic = critic
        assert isinstance(
            self.encoder, NoEncoder
        ), "Define a feature extractor for decoder rather than an encoder in stepwise PPO"

    def evaluate(self, td):
        # Encoder: get encoder output and initial embeddings from initial state
        hidden, _ = self.decoder.feature_extractor(td)
        # pool the embeddings for the critic
        h_tuple = (hidden,) if isinstance(hidden, torch.Tensor) else hidden
        pooled = tuple(map(lambda x: x.mean(dim=-2), h_tuple))
        # potentially cat multiple embeddings (pooled ops and machines)
        h_pooled = torch.cat(pooled, dim=-1)
        # pred value via the value head
        value_pred = self.critic(h_pooled)
        # pre decoder / actor hook
        td, _, hidden = self.decoder.actor.pre_decoder_hook(
            td, None, hidden, num_starts=0
        )
        logits, mask = self.decoder.actor(td, *hidden)
        # get logprobs and entropy over logp distribution
        logprobs = process_logits(logits, mask, tanh_clipping=self.tanh_clipping)
        action_logprobs = gather_by_index(logprobs, td["action"], dim=1)
        dist_entropys = Categorical(logprobs.exp()).entropy()

        return action_logprobs, value_pred, dist_entropys

    def act(self, td, env, phase: str = "train"):
        logits, mask = self.decoder(td, hidden=None, num_starts=0)
        logprobs = process_logits(logits, mask, tanh_clipping=self.tanh_clipping)

        # DRL-S, sampling actions following \pi
        if phase == "train":
            action_indexes = DecodingStrategy.sampling(logprobs)
            td["logprobs"] = gather_by_index(logprobs, action_indexes, dim=1)

        # DRL-G, greedily picking actions with the maximum probability
        else:
            action_indexes = DecodingStrategy.greedy(logprobs)

        # memories.states.append(copy.deepcopy(state))
        td["action"] = action_indexes

        return td

    @torch.no_grad()
    def generate(self, td, env=None, phase: str = "train", **kwargs) -> dict:
        assert phase != "train", "dont use generate() in training mode"
        with torch.no_grad():
            out = super().__call__(td, env, phase=phase, **kwargs)
        return out



from typing import Callable

import torch

from tensordict.tensordict import TensorDict
from torch.distributions import Uniform

from rl4co.envs.common.utils import Generator, get_sampler
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class MPDPGenerator(Generator):
    """Data generator for the Capacitated Vehicle Routing Problem (CVRP).
    Args:
        num_loc: number of locations
        min_loc: minimum location value
        max_loc: maximum location value
        loc_distribution: distribution for the locations
        depot_distribution: distribution for the depot location. If None, sample the depot from the locations
        min_num_agents: minimum number of agents
        max_num_agents: maximum number of agents

    Returns:
        A TensorDict with the following keys:
            locs [batch_size, num_loc, 2]: locations of each customer and the depot
            depot [batch_size, 2]: location of the depot
            num_agents [batch_size]: number of agents
    """

    def __init__(
        self,
        num_loc: int = 20,
        min_loc: float = 0.0,
        max_loc: float = 1.0,
        loc_distribution: int | float | str | type | Callable = Uniform,
        depot_distribution: int | float | str | type | Callable = None,
        min_num_agents: int = 2,
        max_num_agents: int = 10,
        **kwargs,
    ):
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.min_num_agents = min_num_agents
        self.max_num_agents = max_num_agents

        # Number of locations must be even
        if num_loc % 2 != 0:
            log.warn(
                "Number of locations must be even. Adding 1 to the number of locations."
            )
            self.num_loc += 1

        # Location distribution
        if kwargs.get("loc_sampler", None) is not None:
            self.loc_sampler = kwargs["loc_sampler"]
        else:
            self.loc_sampler = get_sampler(
                "loc", loc_distribution, min_loc, max_loc, **kwargs
            )

        # Depot distribution
        if kwargs.get("depot_sampler", None) is not None:
            self.depot_sampler = kwargs["depot_sampler"]
        else:
            self.depot_sampler = (
                get_sampler("depot", depot_distribution, min_loc, max_loc, **kwargs)
                if depot_distribution is not None
                else None
            )

    def _generate(self, batch_size) -> TensorDict:
        # Sample locations: depot and customers
        if self.depot_sampler is not None:
            depot = self.depot_sampler.sample((*batch_size, 2))
            locs = self.loc_sampler.sample((*batch_size, self.num_loc, 2))
        else:
            # if depot_sampler is None, sample the depot from the locations
            locs = self.loc_sampler.sample((*batch_size, self.num_loc + 1, 2))
            depot = locs[..., 0, :]
            locs = locs[..., 1:, :]

        # Sample the number of agents
        num_agents = torch.randint(
            self.min_num_agents,
            self.max_num_agents + 1,
            size=(*batch_size,),
        )

        return TensorDict(
            {
                "locs": locs,
                "depot": depot,
                "num_agents": num_agents,
            },
            batch_size=batch_size,
        )



import numpy as np
import torch
import torch.nn as nn


def pos_init_embedding(pos_name: str, config: dict) -> nn.Module:
    """Get positional embedding. The positional embedding is used for improvement methods to encode current solutions.

    Args:
        pos_name: Positional embeding method name.
        config: A dictionary of configuration options for the initlization.
    """
    embedding_registry = {
        "APE": AbsolutePositionalEmbedding,
        "CPE": CyclicPositionalEmbedding,
    }

    if pos_name not in embedding_registry:
        raise ValueError(
            f"Unknown positional embedding name '{pos_name}'. Available positional embeddings: {embedding_registry.keys()}"
        )

    return embedding_registry[pos_name](**config)


class AbsolutePositionalEmbedding(nn.Module):
    """Absolute Positional Embedding in the original Transformer."""

    def __init__(self, embed_dim):
        super(AbsolutePositionalEmbedding, self).__init__()
        self.embed_dim = embed_dim
        self.pattern = None

    def _init(self, n_position, emb_dim):
        pattern = torch.tensor(
            [
                [pos / np.power(10000, 2 * (j // 2) / emb_dim) for j in range(emb_dim)]
                for pos in range(1, n_position + 1)
            ],
            dtype=torch.float32,
        )

        pattern[1:, 0::2] = torch.sin(pattern[1:, 0::2])  # dim 2i
        pattern[1:, 1::2] = torch.cos(pattern[1:, 1::2])  # dim 2i+1

        return pattern

    def forward(self, td):
        batch_size, seq_length = td["rec_current"].size()
        visited_time = td["visited_time"]
        embedding_dim = self.embed_dim

        # expand for every batch
        if self.pattern is None or self.pattern.size(0) != seq_length:
            self.pattern = self._init(seq_length, self.embed_dim)

        batch_vector = (
            self.pattern.expand(batch_size, seq_length, embedding_dim)
            .clone()
            .to(visited_time.device)
        )
        index = (
            (visited_time % seq_length)
            .long()
            .unsqueeze(-1)
            .expand(batch_size, seq_length, embedding_dim)
        )

        return torch.gather(batch_vector, 1, index)


class CyclicPositionalEmbedding(nn.Module):
    """Cyclic Positional Embedding presented in Ma et al.(2021)
    See https://arxiv.org/abs/2110.02544
    """

    def __init__(self, embed_dim, mean_pooling=True):
        super(CyclicPositionalEmbedding, self).__init__()
        self.embed_dim = embed_dim
        self.mean_pooling = mean_pooling
        self.pattern = None

    def _basesin(self, x, T, fai=0):
        return np.sin(2 * np.pi / T * np.abs(np.mod(x, 2 * T) - T) + fai)

    def _basecos(self, x, T, fai=0):
        return np.cos(2 * np.pi / T * np.abs(np.mod(x, 2 * T) - T) + fai)

    def _init(self, n_position, emb_dim, mean_pooling):
        Td_set = np.linspace(
            np.power(n_position, 1 / (emb_dim // 2)),
            n_position,
            emb_dim // 2,
            dtype="int",
        )
        x = np.zeros((n_position, emb_dim))

        for i in range(emb_dim):
            Td = (
                Td_set[i // 3 * 3 + 1]
                if (i // 3 * 3 + 1) < (emb_dim // 2)
                else Td_set[-1]
            )
            fai = (
                0
                if i <= (emb_dim // 2)
                else 2 * np.pi * ((-i + (emb_dim // 2)) / (emb_dim // 2))
            )
            longer_pattern = np.arange(0, np.ceil((n_position) / Td) * Td, 0.01)
            if i % 2 == 1:
                x[:, i] = self._basecos(longer_pattern, Td, fai)[
                    np.linspace(
                        0, len(longer_pattern), n_position, dtype="int", endpoint=False
                    )
                ]
            else:
                x[:, i] = self._basesin(longer_pattern, Td, fai)[
                    np.linspace(
                        0, len(longer_pattern), n_position, dtype="int", endpoint=False
                    )
                ]

        pattern = torch.from_numpy(x).type(torch.FloatTensor)
        pattern_sum = torch.zeros_like(pattern)

        # averaging the adjacient embeddings if needed (optional, almost the same performance)
        arange = torch.arange(n_position)
        pooling = [0] if not mean_pooling else [-2, -1, 0, 1, 2]
        time = 0
        for i in pooling:
            time += 1
            index = (arange + i + n_position) % n_position
            pattern_sum += pattern.gather(0, index.view(-1, 1).expand_as(pattern))
        pattern = 1.0 / time * pattern_sum - pattern.mean(0)

        return pattern

    def forward(self, td):
        batch_size, seq_length = td["rec_current"].size()
        visited_time = td["visited_time"]
        embedding_dim = self.embed_dim

        # expand for every batch
        if self.pattern is None or self.pattern.size(0) != seq_length:
            self.pattern = self._init(seq_length, self.embed_dim, self.mean_pooling)

        batch_vector = (
            self.pattern.expand(batch_size, seq_length, embedding_dim)
            .clone()
            .to(visited_time.device)
        )
        index = (
            (visited_time % seq_length)
            .long()
            .unsqueeze(-1)
            .expand(batch_size, seq_length, embedding_dim)
        )

        return torch.gather(batch_vector, 1, index)



from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.envs.common.base import ImprovementEnvBase, RL4COEnvBase
from rl4co.utils.ops import gather_by_index, get_tour_length

from .generator import PDPGenerator
from .render import render, render_improvement


class PDPEnv(RL4COEnvBase):
    """Pickup and Delivery Problem (PDP) environment.
    The environment is made of num_loc + 1 locations (cities):
        - 1 depot
        - `num_loc` / 2 pickup locations
        - `num_loc` / 2 delivery locations
    The goal is to visit all the pickup and delivery locations in the shortest path possible starting from the depot
    The conditions is that the agent must visit a pickup location before visiting its corresponding delivery location

    Observations:
        - locations of the depot, pickup, and delivery locations
        - current location of the vehicle
        - the remaining locations to deliver
        - the visited locations
        - the current step

    Constraints:
        - the tour starts and ends at the depot
        - each pickup location must be visited before its corresponding delivery location
        - the vehicle cannot visit the same location twice

    Finish Condition:
        - the vehicle has visited all locations

    Reward:
        - (minus) the negative length of the path

    Args:
        generator: PDPGenerator instance as the data generator
        generator_params: parameters for the generator
        force_start_at_depot: whether to force the agent to start at the depot
            If False (default), the agent won't consider the depot, which is added in the `get_reward` method
            If True, the only valid action at the first step is to visit the depot (=0)
    """

    name = "pdp"

    def __init__(
        self,
        generator: PDPGenerator = None,
        generator_params: dict = {},
        force_start_at_depot: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = PDPGenerator(**generator_params)
        self.generator = generator
        self.force_start_at_depot = force_start_at_depot
        self._make_spec(self.generator)

    @staticmethod
    def _step(td: TensorDict) -> TensorDict:
        current_node = td["action"].unsqueeze(-1)

        num_loc = td["locs"].shape[-2] - 1  # except depot

        # Pickup and delivery node pair of selected node
        new_to_deliver = (current_node + num_loc // 2) % (num_loc + 1)

        # Set available to 0 (i.e., we visited the node)
        available = td["available"].scatter(
            -1, current_node.expand_as(td["action_mask"]), 0
        )

        to_deliver = td["to_deliver"].scatter(
            -1, new_to_deliver.expand_as(td["to_deliver"]), 1
        )

        # Action is feasible if the node is not visited and is to deliver
        # action_mask = torch.logical_and(available, to_deliver)
        action_mask = available & to_deliver

        # We are done there are no unvisited locations
        done = torch.count_nonzero(available, dim=-1) == 0

        # The reward is calculated outside via get_reward for efficiency, so we set it to 0 here
        reward = torch.zeros_like(done)
        
        # Update step
        td.update(
            {
                "current_node": current_node,
                "available": available,
                "to_deliver": to_deliver,
                "i": td["i"] + 1,
                "action_mask": action_mask,
                "reward": reward,
                "done": done,
            }
        )
        return td

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        device = td.device

        locs = torch.cat((td["depot"][:, None, :], td["locs"]), -2)

        # Pick is 1, deliver is 0 [batch_size, graph_size+1], [1,1...1, 0...0]
        to_deliver = torch.cat(
            [
                torch.ones(
                    *batch_size,
                    self.generator.num_loc // 2 + 1,
                    dtype=torch.bool,
                ).to(device),
                torch.zeros(
                    *batch_size,
                    self.generator.num_loc // 2,
                    dtype=torch.bool,
                ).to(device),
            ],
            dim=-1,
        )

        # Masking variables
        available = torch.ones(
            (*batch_size, self.generator.num_loc + 1), dtype=torch.bool
        ).to(device)
        action_mask = torch.ones_like(available) # [batch_size, graph_size+1]
        if self.force_start_at_depot:
            action_mask[..., 1:] = False # can only visit the depot at the first step
        else:
            action_mask = action_mask & to_deliver
            available[..., 0] = False # depot is already visited (during reward calculation)
            action_mask[..., 0] = False # depot is not available to visit

        # Other variables
        current_node = torch.zeros((*batch_size, 1), dtype=torch.int64).to(device)
        i = torch.zeros((*batch_size, 1), dtype=torch.int64).to(device)

        return TensorDict(
            {
                "locs": locs,
                "current_node": current_node,
                "to_deliver": to_deliver,
                "available": available,
                "i": i,
                "action_mask": action_mask,
            },
            batch_size=batch_size,
        )

    def _make_spec(self, generator: PDPGenerator):
        """Make the observation and action specs from the parameters."""
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(generator.num_loc + 1, 2),
                dtype=torch.float32,
            ),
            current_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            to_deliver=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            i=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=Unbounded(
                shape=(generator.num_loc + 1),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=generator.num_loc + 1,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    @staticmethod
    def _get_reward(td, actions) -> TensorDict:
        # Gather locations in order of tour (add depot since we start and end there)
        locs_ordered = torch.cat(
            [
                td["locs"][..., 0:1, :],  # depot
                gather_by_index(td["locs"], actions),  # order locations
            ],
            dim=1,
        )
        return -get_tour_length(locs_ordered)

    def check_solution_validity(self, td, actions):
        if not self.force_start_at_depot:
            actions = torch.cat((torch.zeros_like(actions[:, 0:1]), actions), dim=-1)

        assert (
            (torch.arange(actions.size(1), out=actions.data.new()))
            .view(1, -1)
            .expand_as(actions)
            == actions.data.sort(1)[0]
        ).all(), "Not visiting all nodes"
        
        # make sure we don't go back to the depot in the middle of the tour
        assert (actions[:, 1:-1] != 0).all(), "Going back to depot in the middle of the tour (not allowed)"

        visited_time = torch.argsort(
            actions, 1
        )  # index of pickup less than index of delivery
        assert (
            visited_time[:, 1 : actions.size(1) // 2 + 1]
            < visited_time[:, actions.size(1) // 2 + 1 :]
        ).all(), "Deliverying without pick-up"

    def get_num_starts(self, td):
        """Only half of the nodes (i.e. pickup nodes) can be start nodes"""
        return (td["locs"].shape[-2] - 1) // 2

    def select_start_nodes(self, td, num_starts):
        """Only nodes from [1 : num_loc // 2 +1] (i.e. pickups) can be selected"""
        num_possible_starts = (td["locs"].shape[-2] - 1) // 2
        selected = (
            torch.arange(num_starts, device=td.device).repeat_interleave(td.shape[0])
            % num_possible_starts
            + 1
        )
        return selected

    @staticmethod
    def render(td: TensorDict, actions: torch.Tensor = None, ax=None):
        return render(td, actions, ax)


class PDPRuinRepairEnv(ImprovementEnvBase):
    """Pickup and Delivery Problem (PDP) environment for performing neural ruin-repair search.
    The environment is made of num_loc + 1 locations (cities):
        - 1 depot
        - `num_loc` / 2 pickup locations
        - `num_loc` / 2 delivery locations

    The goal is to search for optimal solutions to pickup and delivery problems by performing a ruin-and-repair neighborhood search on a given initial solution.
    (see MDP described in https://arxiv.org/abs/2204.11399)

    The condition is that at each step, the visited solutions must be feasible,
    maintaining the sequence of visiting the pickup location before its corresponding delivery location.

    Observations:
        - locations of the depot, pickup, and delivery locations
        - current solution to be improved
        - historical decisions
        - the current step

    Constraints:
        - the tour starts and ends at the depot
        - each pickup location must be visited before its corresponding delivery location
        - the vehicle cannot visit the same location twice

    Finish Condition:
        - None

    Reward:
        - the immediate reduced cost over the current best-so-far solution
        (see MDP described in https://arxiv.org/abs/2204.11399)

    Args:
        num_loc: number of locations (cities) in the TSP
        init_sol_type: the method type used for generating initial solutions (random or greedy)
        td_params: parameters of the environment
        seed: seed for the environment
        device: device to use.  Generally, no need to set as tensors are updated on the fly
    """

    name = "pdp_ruin_repair"

    def __init__(
        self,
        generator: PDPGenerator = None,
        generator_params: dict = {},
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = PDPGenerator(**generator_params)
        self.generator = generator
        self._make_spec(self.generator)

    def _step(self, td: TensorDict, solution_to=None) -> TensorDict:
        # get state information from td
        solution_best = td["rec_best"]
        locs = td["locs"]
        cost_bsf = td["cost_bsf"]
        action_record = td["action_record"]
        bs, gs = solution_best.size()

        # perform local_operator
        if solution_to is None:
            action = td["action"]
            solution = td["rec_current"]
            next_rec = self._local_operator(solution, action)
        else:
            next_rec = solution_to.clone()
        new_obj = self.get_costs(locs, next_rec)

        # compute reward and update best-so-far solutions
        now_bsf = torch.where(new_obj < cost_bsf, new_obj, cost_bsf)
        reward = cost_bsf - now_bsf
        index = reward > 0.0
        solution_best[index] = next_rec[index].clone()

        # reset visited_time
        visited_time = td["visited_time"] * 0
        pre = torch.zeros((bs), device=visited_time.device).long()
        arange = torch.arange(bs)
        for i in range(gs):
            current_nodes = next_rec[arange, pre]
            visited_time[arange, current_nodes] = i + 1
            pre = current_nodes
        visited_time = visited_time.long()

        # update action record
        if solution_to is None:
            action_record[:, :-1] = action_record[:, 1:]
            action_record[:, -1] *= 0
            action_record[torch.arange(bs), -1, action[:, 0]] = 1

        # Update step
        td.update(
            {
                "cost_current": new_obj,
                "cost_bsf": now_bsf,
                "rec_current": next_rec,
                "rec_best": solution_best,
                "visited_time": visited_time,
                "action_record": action_record,
                "i": td["i"] + 1 if solution_to is None else td["i"],
                "reward": reward,
            }
        )

        return td

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        device = td.device

        locs = torch.cat((td["depot"][:, None, :], td["locs"]), -2)
        current_rec = self.generator._get_initial_solutions(locs).to(device)
        obj = self.get_costs(locs, current_rec)

        # get index according to the solutions in the linked list data structure
        bs = batch_size[0]
        seq_length = self.generator.num_loc + 1
        visited_time = torch.zeros((bs, seq_length)).to(device)
        pre = torch.zeros((bs)).to(device).long()
        arange = torch.arange(bs)
        for i in range(seq_length):
            current_nodes = current_rec[arange, pre]
            visited_time[arange, current_nodes] = i + 1
            pre = current_nodes
        visited_time = visited_time.long()

        # get action record and step i
        i = torch.zeros((*batch_size, 1), dtype=torch.int64).to(device)
        action_record = (
            torch.zeros((bs, seq_length, seq_length // 2))
            if self.training
            else torch.zeros((bs, seq_length // 2, seq_length // 2))
        )

        return TensorDict(
            {
                "locs": locs,
                "cost_current": obj,
                "cost_bsf": obj.clone(),
                "rec_current": current_rec,
                "rec_best": current_rec.clone(),
                "visited_time": visited_time,
                "action_record": action_record,
                "i": i,
            },
            batch_size=batch_size,
        )

    @staticmethod
    def _local_operator(solution, action):
        # get info
        pair_index = action[:, 0].view(-1, 1) + 1
        first = action[:, 1].view(-1, 1)
        second = action[:, 2].view(-1, 1)
        rec = solution.clone()
        bs, gs = rec.size()

        # fix connection for pairing node
        argsort = rec.argsort()
        pre_pairfirst = argsort.gather(1, pair_index)
        post_pairfirst = rec.gather(1, pair_index)
        rec.scatter_(1, pre_pairfirst, post_pairfirst)
        rec.scatter_(1, pair_index, pair_index)

        argsort = rec.argsort()

        pre_pairsecond = argsort.gather(1, pair_index + gs // 2)
        post_pairsecond = rec.gather(1, pair_index + gs // 2)

        rec.scatter_(1, pre_pairsecond, post_pairsecond)

        # fix connection for pairing node
        post_second = rec.gather(1, second)
        rec.scatter_(1, second, pair_index + gs // 2)
        rec.scatter_(1, pair_index + gs // 2, post_second)

        post_first = rec.gather(1, first)
        rec.scatter_(1, first, pair_index)
        rec.scatter_(1, pair_index, post_first)

        return rec

    def _make_spec(self, generator: PDPGenerator):
        """Make the observation and action specs from the parameters."""
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(generator.num_loc + 1, 2),
                dtype=torch.float32,
            ),
            cost_current=Unbounded(
                shape=(1),
                dtype=torch.float32,
            ),
            cost_bsf=Unbounded(
                shape=(1),
                dtype=torch.float32,
            ),
            rec_current=Unbounded(
                shape=(self.generator.num_loc + 1),
                dtype=torch.int64,
            ),
            rec_best=Unbounded(
                shape=(self.generator.num_loc + 1),
                dtype=torch.int64,
            ),
            visited_time=Unbounded(
                shape=(self.generator.num_loc + 1, self.generator.num_loc + 1),
                dtype=torch.int64,
            ),
            action_record=Unbounded(
                shape=(self.generator.num_loc + 1, self.generator.num_loc + 1),
                dtype=torch.int64,
            ),
            i=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(3,),
            dtype=torch.int64,
            low=0,
            high=self.generator.num_loc + 1,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    def check_solution_validity(self, td, actions=None):
        # The function can be called by the agent to check the validity of the best found solution
        # Note that the args actions are not used in improvement method.

        solution = td["rec_best"]
        batch_size, graph_size = solution.size()

        assert (
            torch.arange(graph_size, out=solution.data.new())
            .view(1, -1)
            .expand_as(solution)
            == solution.data.sort(1)[0]
        ).all(), "Not visiting all nodes"

        visited_time = torch.zeros((batch_size, graph_size), device=self.device)
        pre = torch.zeros(batch_size, device=self.device).long()
        arange = torch.arange(batch_size)
        for i in range(graph_size):
            visited_time[arange, solution[arange, pre]] = i + 1
            pre = solution[arange, pre]

        assert (
            visited_time[:, 1 : graph_size // 2 + 1]
            < visited_time[:, graph_size // 2 + 1 :]
        ).all(), "Deliverying without pick-up"

    @staticmethod
    def get_mask(selected_node, td):
        # return mask that is 1 if the corresponding action is feasible, 0 otherwise

        visited_time = td["visited_time"]
        bs, gs = visited_time.size()
        visited_time = visited_time % gs
        arange = torch.arange(bs)

        visited_order_map = visited_time.view(bs, gs, 1) > visited_time.view(bs, 1, gs)
        mask = visited_order_map.clone()
        mask[arange, selected_node.view(-1)] = True
        mask[arange, selected_node.view(-1) + gs // 2] = True
        mask[arange, :, selected_node.view(-1)] = True
        mask[arange, :, selected_node.view(-1) + gs // 2] = True

        bs, gs, _ = visited_order_map.size()

        return ~mask

    @classmethod
    def _random_action(cls, td):
        batch_size, graph_size = td["rec_best"].size()
        selected_node = (
            (torch.rand(batch_size, 1) * graph_size // 2) % (graph_size // 2)
        ).long()
        mask = cls.get_mask(selected_node + 1, td)
        logits = torch.rand(batch_size, graph_size, graph_size)
        logits[~mask] = -1e20
        prob = torch.softmax(logits.view(batch_size, -1), -1)
        sample = prob.multinomial(1)
        action = torch.cat(
            (selected_node, sample // (graph_size), sample % (graph_size)), -1
        )
        td["action"] = action
        return action

    @classmethod
    def render(cls, td: TensorDict, actions: torch.Tensor = None, ax=None):
        solution_current = cls.get_current_solution(td)
        solution_best = cls.get_best_solution(td)
        return render_improvement(td, solution_current, solution_best)



from rl4co.models.nn.env_embeddings.init import JSSPInitEmbedding
from rl4co.models.nn.graph.gcn import GCNEncoder
from rl4co.utils.ops import adj_to_pyg_edge_index


class GCN4JSSP(GCNEncoder):
    def __init__(
        self,
        embed_dim: int,
        num_layers: int,
        init_embedding=None,
        **init_embedding_kwargs,
    ):
        def edge_idx_fn(td, _):
            return adj_to_pyg_edge_index(td["adjacency"])

        if init_embedding is None:
            init_embedding = JSSPInitEmbedding(embed_dim, **init_embedding_kwargs)

        super().__init__(
            env_name="jssp",
            embed_dim=embed_dim,
            num_layers=num_layers,
            edge_idx_fn=edge_idx_fn,
            init_embedding=init_embedding,
        )



from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.pylogger import get_pylogger

from .generator import SMTWTPGenerator
from .render import render

log = get_pylogger(__name__)


class SMTWTPEnv(RL4COEnvBase):
    """
    Single Machine Total Weighted Tardiness Problem environment as described in DeepACO (https://arxiv.org/pdf/2309.14032.pdf)
    SMTWTP is a scheduling problem in which a set of jobs must be processed on a single machine.
    Each job i has a processing time, a weight, and a due date. The objective is to minimize the sum of the weighted tardiness of all jobs,
    where the weighted tardiness of a job is defined as the product of its weight and the duration by which its completion time exceeds its due date.
    At each step, the agent chooses a job to process. The reward is 0 unless the agent processes all the jobs.
    In that case, the reward is (-)objective value of the processing order: maximizing the reward is equivalent to minimizing the objective.

    Observation:
        - job_due_time: the due time of each job
        - job_weight: the weight of each job
        - job_process_time: the process time of each job
        - current_node: the current node
        - action_mask: a mask of available actions
        - current_time: the current time

    Constants:
        - num_job: number of jobs
        - min_time_span: lower bound of jobs' due time. By default, jobs' due time is uniformly sampled from (min_time_span, max_time_span)
        - max_time_span: upper bound of jobs' due time. By default, it will be set to num_job / 2
        - min_job_weight: lower bound of jobs' weights. By default, jobs' weights are uniformly sampled from (min_job_weight, max_job_weight)
        - max_job_weight: upper bound of jobs' weights
        - min_process_time: lower bound of jobs' process time. By default, jobs' process time is uniformly sampled from (min_process_time, max_process_time)
        - max_process_time: upper bound of jobs' process time

    Finishing condition:
        - All jobs are processed

    Reward:
        - The reward is 0 unless the agent processes all the jobs.
        - In that case, the reward is (-)objective value of the processing order: maximizing the reward is equivalent to minimizing the objective.

    Args:
        generator: FFSPGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "smtwtp"

    def __init__(
        self,
        generator: SMTWTPGenerator = None,
        generator_params: dict = {},
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = SMTWTPGenerator(**generator_params)
        self.generator = generator
        self._make_spec(self.generator)

    @staticmethod
    def _step(td: TensorDict) -> TensorDict:
        current_job = td["action"]

        # Set not visited to 0 (i.e., we visited the node)
        available = td["action_mask"].scatter(
            -1, current_job.unsqueeze(-1).expand_as(td["action_mask"]), 0
        )

        # Increase used time
        selected_process_time = td["job_process_time"][
            torch.arange(current_job.size(0)), current_job
        ]
        current_time = td["current_time"] + selected_process_time.unsqueeze(-1)

        # We are done there are no unvisited locations
        done = torch.count_nonzero(available, dim=-1) <= 0

        # The reward is calculated outside via get_reward for efficiency, so we set it to 0 here
        reward = torch.zeros_like(done)

        td.update(
            {
                "current_job": current_job,
                "current_time": current_time,
                "action_mask": available,
                "reward": reward,
                "done": done,
            }
        )
        return td

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        device = td.device

        init_job_due_time = td["job_due_time"]
        init_job_process_time = td["job_process_time"]
        init_job_weight = td["job_weight"]

        # Other variables
        current_job = torch.zeros((*batch_size, 1), dtype=torch.int64, device=device)
        current_time = torch.zeros((*batch_size, 1), dtype=torch.int64, device=device)
        available = torch.ones(
            (*batch_size, self.generator.num_job + 1), dtype=torch.bool, device=device
        )
        available[:, 0] = 0  # mask the starting dummy node

        return TensorDict(
            {
                "job_due_time": init_job_due_time,
                "job_weight": init_job_weight,
                "job_process_time": init_job_process_time,
                "current_job": current_job,
                "current_time": current_time,
                "action_mask": available,
            },
            batch_size=batch_size,
        )

    def _make_spec(self, generator: SMTWTPGenerator) -> None:
        self.observation_spec = Composite(
            job_due_time=Bounded(
                low=generator.min_time_span,
                high=generator.max_time_span,
                shape=(generator.num_job + 1,),
                dtype=torch.float32,
            ),
            job_weight=Bounded(
                low=generator.min_job_weight,
                high=generator.max_job_weight,
                shape=(generator.num_job + 1,),
                dtype=torch.float32,
            ),
            job_process_time=Bounded(
                low=generator.min_process_time,
                high=generator.max_process_time,
                shape=(generator.num_job + 1,),
                dtype=torch.float32,
            ),
            current_node=Unbounded(
                shape=(1,),
                dtype=torch.int64,
            ),
            action_mask=Unbounded(
                shape=(generator.num_job + 1,),
                dtype=torch.bool,
            ),
            current_time=Unbounded(
                shape=(1,),
                dtype=torch.float32,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=generator.num_job + 1,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    def _get_reward(self, td, actions) -> TensorDict:
        job_due_time = td["job_due_time"]
        job_weight = td["job_weight"]
        job_process_time = td["job_process_time"]

        batch_idx = torch.arange(
            job_process_time.shape[0], device=job_process_time.device
        ).unsqueeze(1)

        ordered_process_time = job_process_time[batch_idx, actions]
        ordered_due_time = job_due_time[batch_idx, actions]
        ordered_job_weight = job_weight[batch_idx, actions]
        presum_process_time = torch.cumsum(
            ordered_process_time, dim=1
        )  # ending time of each job
        job_tardiness = presum_process_time - ordered_due_time
        job_tardiness[job_tardiness < 0] = 0
        job_weighted_tardiness = ordered_job_weight * job_tardiness

        return -job_weighted_tardiness.sum(-1)

    def check_solution_validity(self, td, actions):
        log.warning("Checking solution validity is not implemented for SMTWTP")
        pass

    @staticmethod
    def render(td, actions=None, ax=None):
        return render(td, actions, ax)



import matplotlib.pyplot as plt
import torch

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


# Only render the first instance
def render(td, actions=None, ax=None):
    markersize = 8

    td = td.detach().cpu()
    # if batch_size greater than 0 , we need to select the first batch element
    if td.batch_size != torch.Size([]):
        td = td[0]
        if actions is not None:
            actions = actions[0]

    # Variables
    init_deliveries = td["to_deliver"][1:]
    delivery_locs = td["locs"][1:][~init_deliveries.bool()]
    pickup_locs = td["locs"][1:][init_deliveries.bool()]
    depot_loc = td["locs"][0]
    actions = actions if actions is not None else td["action"]

    fig, ax = plt.subplots()

    # Plot the actions in order
    for i in range(len(actions)):
        from_node = actions[i]
        to_node = (
            actions[i + 1] if i < len(actions) - 1 else actions[0]
        )  # last goes back to depot
        from_loc = td["locs"][from_node]
        to_loc = td["locs"][to_node]
        ax.plot([from_loc[0], to_loc[0]], [from_loc[1], to_loc[1]], "k-")
        ax.annotate(
            "",
            xy=(to_loc[0], to_loc[1]),
            xytext=(from_loc[0], from_loc[1]),
            arrowprops=dict(arrowstyle="->", color="black"),
            annotation_clip=False,
        )

    # Plot the depot location
    ax.plot(
        depot_loc[0],
        depot_loc[1],
        "g",
        marker="s",
        markersize=markersize,
        label="Depot",
    )

    # Plot the pickup locations
    for i, pickup_loc in enumerate(pickup_locs):
        ax.plot(
            pickup_loc[0],
            pickup_loc[1],
            "r",
            marker="^",
            markersize=markersize,
            label="Pickup" if i == 0 else None,
        )

    # Plot the delivery locations
    for i, delivery_loc in enumerate(delivery_locs):
        ax.plot(
            delivery_loc[0],
            delivery_loc[1],
            "b",
            marker="v",
            markersize=markersize,
            label="Delivery" if i == 0 else None,
        )

    return ax


def render_improvement(td, current_soltuion, best_soltuion):
    coordinates = td["locs"][0]
    real_seq = current_soltuion[:1]
    real_best = best_soltuion[:1]
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 6))  # Create two side-by-side axes

    for ax in [ax1, ax2]:  # Plot on both axes
        if ax == ax1:
            ax.axis([-0.05, 1.05] * 2)
            # plot the nodes
            ax.scatter(
                coordinates[:, 0], coordinates[:, 1], marker="H", s=55, c="blue", zorder=2
            )
            # plot the tour
            real_seq_coordinates = coordinates.gather(
                0, real_seq[0].unsqueeze(1).repeat(1, 2)
            )
            real_seq_coordinates = torch.cat(
                (real_seq_coordinates, real_seq_coordinates[:1]), 0
            )
            ax.plot(
                real_seq_coordinates[:, 0],
                real_seq_coordinates[:, 1],
                color="black",
                zorder=1,
            )
            # mark node
            for i, txt in enumerate(range(real_seq.size(1))):
                ax.annotate(
                    txt,
                    (coordinates[i, 0] + 0.01, coordinates[i, 1] + 0.01),
                )
            ax.set_title("Current Solution")
        else:
            ax.axis([-0.05, 1.05] * 2)
            # plot the nodes
            ax.scatter(
                coordinates[:, 0], coordinates[:, 1], marker="H", s=55, c="blue", zorder=2
            )
            # plot the tour
            real_best_coordinates = coordinates.gather(
                0, real_best[0].unsqueeze(1).repeat(1, 2)
            )
            real_best_coordinates = torch.cat(
                (real_best_coordinates, real_best_coordinates[:1]), 0
            )
            ax.plot(
                real_best_coordinates[:, 0],
                real_best_coordinates[:, 1],
                color="black",
                zorder=1,
            )
            # mark node
            for i, txt in enumerate(range(real_seq.size(1))):
                ax.annotate(
                    txt,
                    (coordinates[i, 0] + 0.01, coordinates[i, 1] + 0.01),
                )
            ax.set_title("Best Solution")
    plt.tight_layout()

    return ax



from typing import Callable, Optional

import torch.nn as nn

from torch import Tensor

from rl4co.models.nn.mlp import MLP
from rl4co.models.nn.moe import MoE
from rl4co.models.nn.attention import MultiHeadAttention
from rl4co.models.nn.ops import Normalization, SkipConnection
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class MultiHeadAttentionLayer(nn.Sequential):
    """Multi-Head Attention Layer with normalization and feed-forward layer

    Args:
        embed_dim: dimension of the embeddings
        num_heads: number of heads in the MHA
        feedforward_hidden: dimension of the hidden layer in the feed-forward layer
        normalization: type of normalization to use (batch, layer, none)
        sdpa_fn: scaled dot product attention function (SDPA)
        moe_kwargs: Keyword arguments for MoE
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int = 8,
        feedforward_hidden: int = 512,
        normalization: Optional[str] = "batch",
        bias: bool = True,
        sdpa_fn: Optional[Callable] = None,
        moe_kwargs: Optional[dict] = None,
    ):
        num_neurons = [feedforward_hidden] if feedforward_hidden > 0 else []
        if moe_kwargs is not None:
            ffn = MoE(embed_dim, embed_dim, num_neurons=num_neurons, **moe_kwargs)
        else:
            ffn = MLP(input_dim=embed_dim, output_dim=embed_dim, num_neurons=num_neurons, hidden_act="ReLU")

        super(MultiHeadAttentionLayer, self).__init__(
            SkipConnection(
                MultiHeadAttention(embed_dim, num_heads, bias=bias, sdpa_fn=sdpa_fn)
            ),
            Normalization(embed_dim, normalization),
            SkipConnection(ffn),
            Normalization(embed_dim, normalization),
        )


class GraphAttentionNetwork(nn.Module):
    """Graph Attention Network to encode embeddings with a series of MHA layers consisting of a MHA layer,
    normalization, feed-forward layer, and normalization. Similar to Transformer encoder, as used in Kool et al. (2019).

    Args:
        num_heads: number of heads in the MHA
        embed_dim: dimension of the embeddings
        num_layers: number of MHA layers
        normalization: type of normalization to use (batch, layer, none)
        feedforward_hidden: dimension of the hidden layer in the feed-forward layer
        sdpa_fn: scaled dot product attention function (SDPA)
        moe_kwargs: Keyword arguments for MoE
    """

    def __init__(
        self,
        num_heads: int,
        embed_dim: int,
        num_layers: int,
        normalization: str = "batch",
        feedforward_hidden: int = 512,
        sdpa_fn: Optional[Callable] = None,
        moe_kwargs: Optional[dict] = None,
    ):
        super(GraphAttentionNetwork, self).__init__()

        self.layers = nn.Sequential(
            *(
                MultiHeadAttentionLayer(
                    embed_dim,
                    num_heads,
                    feedforward_hidden=feedforward_hidden,
                    normalization=normalization,
                    sdpa_fn=sdpa_fn,
                    moe_kwargs=moe_kwargs,
                )
                for _ in range(num_layers)
            )
        )

    def forward(self, x: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        """Forward pass of the encoder

        Args:
            x: [batch_size, graph_size, embed_dim] initial embeddings to process
            mask: [batch_size, graph_size, graph_size] mask for the input embeddings. Unused for now.
        """
        assert mask is None, "Mask not yet supported!"
        h = self.layers(x)
        return h



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from rl4co.models.nn.attention import MultiHeadCrossAttention
from rl4co.models.nn.env_embeddings import env_init_embedding
from rl4co.models.nn.ops import TransformerFFN


class MixedScoresSDPA(nn.Module):
    def __init__(
        self,
        num_heads: int,
        num_scores: int = 1,
        mixer_hidden_dim: int = 16,
        mix1_init: float = (1 / 2) ** (1 / 2),
        mix2_init: float = (1 / 16) ** (1 / 2),
    ):
        super().__init__()
        self.num_heads = num_heads
        self.num_scores = num_scores
        mix_W1 = torch.torch.distributions.Uniform(low=-mix1_init, high=mix1_init).sample(
            (num_heads, self.num_scores + 1, mixer_hidden_dim)
        )
        mix_b1 = torch.torch.distributions.Uniform(low=-mix1_init, high=mix1_init).sample(
            (num_heads, mixer_hidden_dim)
        )
        self.mix_W1 = nn.Parameter(mix_W1)
        self.mix_b1 = nn.Parameter(mix_b1)

        mix_W2 = torch.torch.distributions.Uniform(low=-mix2_init, high=mix2_init).sample(
            (num_heads, mixer_hidden_dim, 1)
        )
        mix_b2 = torch.torch.distributions.Uniform(low=-mix2_init, high=mix2_init).sample(
            (num_heads, 1)
        )
        self.mix_W2 = nn.Parameter(mix_W2)
        self.mix_b2 = nn.Parameter(mix_b2)

    def forward(self, q, k, v, attn_mask=None, dmat=None, dropout_p=0.0):
        """Scaled Dot-Product Attention with MatNet Scores Mixer"""
        assert dmat is not None
        b, m, n = dmat.shape[:3]
        dmat = dmat.reshape(b, m, n, self.num_scores)

        # Calculate scaled dot product
        attn_scores = torch.matmul(q, k.transpose(-2, -1)) / (k.size(-1) ** 0.5)
        # [b, h, m, n, num_scores+1]
        mix_attn_scores = torch.cat(
            [
                attn_scores.unsqueeze(-1),
                dmat[:, None, ...].expand(b, self.num_heads, m, n, self.num_scores),
            ],
            dim=-1,
        )
        # [b, h, m, n]
        attn_scores = (
            (
                torch.matmul(
                    F.relu(
                        torch.matmul(mix_attn_scores.transpose(1, 2), self.mix_W1)
                        + self.mix_b1[None, None, :, None, :]
                    ),
                    self.mix_W2,
                )
                + self.mix_b2[None, None, :, None, :]
            )
            .transpose(1, 2)
            .squeeze(-1)
        )

        # Apply the provided attention mask
        if attn_mask is not None:
            if attn_mask.dtype == torch.bool:
                attn_mask[~attn_mask.any(-1)] = True
                attn_scores.masked_fill_(~attn_mask, float("-inf"))
            else:
                attn_scores += attn_mask

        # Softmax to get attention weights
        attn_weights = F.softmax(attn_scores, dim=-1)

        # Apply dropout
        if dropout_p > 0.0:
            attn_weights = F.dropout(attn_weights, p=dropout_p)

        # Compute the weighted sum of values
        return torch.matmul(attn_weights, v)


class MatNetCrossMHA(MultiHeadCrossAttention):
    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        bias: bool = False,
        mixer_hidden_dim: int = 16,
        mix1_init: float = (1 / 2) ** (1 / 2),
        mix2_init: float = (1 / 16) ** (1 / 2),
    ):
        attn_fn = MixedScoresSDPA(
            num_heads=num_heads,
            mixer_hidden_dim=mixer_hidden_dim,
            mix1_init=mix1_init,
            mix2_init=mix2_init,
        )

        super().__init__(
            embed_dim=embed_dim, num_heads=num_heads, bias=bias, sdpa_fn=attn_fn
        )


class MatNetMHA(nn.Module):
    def __init__(self, embed_dim: int, num_heads: int, bias: bool = False):
        super().__init__()
        self.row_encoding_block = MatNetCrossMHA(embed_dim, num_heads, bias)
        self.col_encoding_block = MatNetCrossMHA(embed_dim, num_heads, bias)

    def forward(self, row_emb, col_emb, dmat, attn_mask=None):
        """
        Args:
            row_emb (Tensor): [b, m, d]
            col_emb (Tensor): [b, n, d]
            dmat (Tensor): [b, m, n]

        Returns:
            Updated row_emb (Tensor): [b, m, d]
            Updated col_emb (Tensor): [b, n, d]
        """
        updated_row_emb = self.row_encoding_block(
            row_emb, col_emb, dmat=dmat, cross_attn_mask=attn_mask
        )
        attn_mask_t = attn_mask.transpose(-2, -1) if attn_mask is not None else None
        updated_col_emb = self.col_encoding_block(
            col_emb,
            row_emb,
            dmat=dmat.transpose(-2, -1),
            cross_attn_mask=attn_mask_t,
        )
        return updated_row_emb, updated_col_emb


class MatNetLayer(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        bias: bool = False,
        feedforward_hidden: int = 512,
        normalization: Optional[str] = "instance",
    ):
        super().__init__()
        self.MHA = MatNetMHA(embed_dim, num_heads, bias)
        self.F_a = TransformerFFN(embed_dim, feedforward_hidden, normalization)
        self.F_b = TransformerFFN(embed_dim, feedforward_hidden, normalization)

    def forward(self, row_emb, col_emb, dmat, attn_mask=None):
        """
        Args:
            row_emb (Tensor): [b, m, d]
            col_emb (Tensor): [b, n, d]
            dmat (Tensor): [b, m, n]

        Returns:
            Updated row_emb (Tensor): [b, m, d]
            Updated col_emb (Tensor): [b, n, d]
        """

        row_emb_out, col_emb_out = self.MHA(row_emb, col_emb, dmat, attn_mask)
        row_emb_out = self.F_a(row_emb_out, row_emb)
        col_emb_out = self.F_b(col_emb_out, col_emb)
        return row_emb_out, col_emb_out


class MatNetEncoder(nn.Module):
    def __init__(
        self,
        embed_dim: int = 256,
        num_heads: int = 16,
        num_layers: int = 3,
        normalization: str = "batch",
        feedforward_hidden: int = 512,
        init_embedding: nn.Module = None,
        init_embedding_kwargs: dict = {},
        bias: bool = False,
        mask_non_neighbors: bool = False,
    ):
        super().__init__()

        if init_embedding is None:
            init_embedding = env_init_embedding(
                "matnet", {"embed_dim": embed_dim, **init_embedding_kwargs}
            )

        self.init_embedding = init_embedding
        self.mask_non_neighbors = mask_non_neighbors
        self.layers = nn.ModuleList(
            [
                MatNetLayer(
                    embed_dim=embed_dim,
                    num_heads=num_heads,
                    bias=bias,
                    feedforward_hidden=feedforward_hidden,
                    normalization=normalization,
                )
                for _ in range(num_layers)
            ]
        )

    def forward(self, td, attn_mask: torch.Tensor = None):
        row_emb, col_emb, dmat = self.init_embedding(td)

        if self.mask_non_neighbors and attn_mask is None:
            # attn_mask (keep 1s discard 0s) to only attend on neighborhood
            attn_mask = dmat.ne(0)

        for layer in self.layers:
            row_emb, col_emb = layer(row_emb, col_emb, dmat, attn_mask)

        embedding = (row_emb, col_emb)
        init_embedding = None
        return embedding, init_embedding  # match output signature for the AR policy class



from typing import Optional

import torch
import torch.nn.functional as F

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.ops import gather_by_index, get_tour_length
from rl4co.utils.pylogger import get_pylogger

from .generator import PCTSPGenerator
from .render import render

log = get_pylogger(__name__)


class PCTSPEnv(RL4COEnvBase):
    """Prize-collecting TSP (PCTSP) environment.
    The goal is to collect as much prize as possible while minimizing the total travel cost.
    The environment is stochastic, the prize is only revealed when the node is visited.

    Observations:
        - locations of the nodes
        - prize and penalty of each node
        - current location of the vehicle
        - current total prize
        - current total penalty
        - visited nodes
        - prize required to visit a node
        - the current step

    Constraints:
        - the tour starts and ends at the depot
        - the vehicle cannot visit nodes exceed the remaining prize

    Finish Condition:
        - the vehicle back to the depot

    Reward:
        - the sum of the saved penalties

    Args:
        generator: OPGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "pctsp"
    _stochastic = False

    def __init__(
        self,
        generator: PCTSPGenerator = None,
        generator_params: dict = {},
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = PCTSPGenerator(**generator_params)
        self.generator = generator
        self._make_spec(self.generator)

    def _step(self, td: TensorDict) -> TensorDict:
        current_node = td["action"]

        # Get current coordinates, prize, and penalty
        cur_total_prize = td["cur_total_prize"] + gather_by_index(
            td["real_prize"], current_node
        )
        cur_total_penalty = td["cur_total_penalty"] + gather_by_index(
            td["penalty"], current_node
        )

        # Update visited
        visited = td["visited"].scatter(-1, current_node[..., None], 1)

        # Done and reward
        done = (td["i"] > 0) & (current_node == 0)

        # The reward is calculated outside via get_reward for efficiency, so we set it to 0 here
        reward = torch.zeros_like(done)

        # Update state
        td.update(
            {
                "current_node": current_node,
                "cur_total_prize": cur_total_prize,
                "cur_total_penalty": cur_total_penalty,
                "visited": visited,
                "i": td["i"] + 1,
                "reward": reward,
                "done": done,
            }
        )
        td.set("action_mask", self.get_action_mask(td))
        return td

    def _reset(
        self, td: Optional[TensorDict] = None, batch_size: Optional[list] = None
    ) -> TensorDict:
        device = td.device

        locs = torch.cat([td["depot"][..., None, :], td["locs"]], dim=-2)
        expected_prize = td["deterministic_prize"]
        real_prize = (
            td["stochastic_prize"] if self.stochastic else td["deterministic_prize"]
        )
        penalty = td["penalty"]

        # Concatenate depots
        real_prize_with_depot = torch.cat(
            [torch.zeros_like(real_prize[..., :1]), real_prize], dim=-1
        )
        penalty_with_depot = F.pad(penalty, (1, 0), mode="constant", value=0)

        # Initialize the current node and  prize / penalty
        current_node = torch.zeros((*batch_size,), dtype=torch.int64, device=device)
        cur_total_prize = torch.zeros(*batch_size, device=device)
        cur_total_penalty = penalty.sum(-1)  # Sum penalties (all when nothing is visited)

        # Init the action mask (all nodes are available)
        visited = torch.zeros(
            (*batch_size, self.generator.num_loc + 1), dtype=torch.bool, device=device
        )
        i = torch.zeros((*batch_size,), dtype=torch.int64, device=device)
        prize_required = torch.full(
            (*batch_size,), self.generator.prize_required, device=device
        )

        td_reset = TensorDict(
            {
                "locs": locs,
                "current_node": current_node,
                "expected_prize": expected_prize,
                "real_prize": real_prize_with_depot,
                "penalty": penalty_with_depot,
                "cur_total_prize": cur_total_prize,
                "cur_total_penalty": cur_total_penalty,
                "visited": visited,
                "prize_required": prize_required,
                "i": i,
            },
            batch_size=batch_size,
        )
        td_reset.set("action_mask", self.get_action_mask(td_reset))
        return td_reset

    @staticmethod
    def get_action_mask(td: TensorDict) -> torch.Tensor:
        """Cannot visit depot if not yet collected 1 total prize and there are unvisited nodes"""
        mask = td["visited"] | td["visited"][..., 0:1]
        mask[..., 0] = (td["cur_total_prize"] < 1.0) & (
            td["visited"][..., 1:].int().sum(-1) < td["visited"][..., 1:].size(-1)
        )
        return ~(mask > 0)  # Invert mask, since 1 means feasible action

    def _get_reward(self, td: TensorDict, actions: torch.Tensor) -> torch.Tensor:
        """Reward is `saved penalties - (total length + penalty)`"""

        # In case all tours directly return to depot, prevent further problems
        if actions.size(-1) == 1:
            assert (actions == 0).all(), "If all length 1 tours, they should be zero"
            return torch.zeros(actions.size(0), dtype=torch.float, device=actions.device)

        # Gather locations in order of tour (add depot since we start and end there)
        locs_ordered = torch.cat(
            [
                td["locs"][..., 0:1, :],  # depot
                gather_by_index(td["locs"], actions),  # order locations
            ],
            dim=1,
        )
        length = get_tour_length(locs_ordered)

        # Reward is saved penalties - (total length + penalty)
        saved_penalty = td["penalty"].gather(1, actions)
        return saved_penalty.sum(-1) - (length + td["penalty"][..., 1:].sum(-1))

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor) -> None:
        """Check that the solution is valid, i.e. contains all nodes once at most, and either prize constraint is met or all nodes are visited"""

        # Check that tours are valid, i.e. contain 0 to n -1
        sorted_actions = actions.data.sort(1)[0]

        # Make sure each node visited once at most (except for depot)
        assert (
            (sorted_actions[..., 1:] == 0)
            | (sorted_actions[..., 1:] > sorted_actions[..., :-1])
        ).all(), "Duplicates"

        prize = td["real_prize"][..., 1:]  # Remove depot
        prize_with_depot = torch.cat((torch.zeros_like(prize[:, :1]), prize), 1)
        p = prize_with_depot.gather(1, actions)

        # Either prize constraint should be satisfied or all prizes should be visited
        assert (
            (p.sum(-1) >= 1 - 1e-5)
            | (
                sorted_actions.size(-1) - (sorted_actions == 0).int().sum(-1)
                == (td["locs"].size(-2) - 1)
            )  # no depot
        ).all(), "Total prize does not satisfy min total prize"

    @property
    def stochastic(self):
        return self._stochastic

    @stochastic.setter
    def stochastic(self, state: bool):
        if state is True:
            log.warning(
                "Stochastic mode should not be used for PCTSP. Use SPCTSP instead."
            )

    def _make_spec(self, generator):
        """Make the locs and action specs from the parameters."""
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(generator.num_loc, 2),
                dtype=torch.float32,
            ),
            current_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            expected_prize=Unbounded(
                shape=(generator.num_loc),
                dtype=torch.float32,
            ),
            real_prize=Unbounded(
                shape=(generator.num_loc + 1),
                dtype=torch.float32,
            ),
            penalty=Unbounded(
                shape=(generator.num_loc + 1),
                dtype=torch.float32,
            ),
            cur_total_prize=Unbounded(
                shape=(1),
                dtype=torch.float32,
            ),
            cur_total_penalty=Unbounded(
                shape=(1),
                dtype=torch.float32,
            ),
            visited=Unbounded(
                shape=(generator.num_loc + 1),
                dtype=torch.bool,
            ),
            prize_required=Unbounded(
                shape=(1),
                dtype=torch.float32,
            ),
            i=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=Unbounded(
                shape=(generator.num_loc),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=generator.num_loc,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    @staticmethod
    def render(td: TensorDict, actions: torch.Tensor = None, ax=None):
        return render(td, actions, ax)



import matplotlib.pyplot as plt
import numpy as np
import torch

from matplotlib import colormaps

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def render(td, actions=None, ax=None):
    def discrete_cmap(num, base_cmap="nipy_spectral"):
        """Create an N-bin discrete colormap from the specified input map"""
        base = colormaps[base_cmap]
        color_list = base(np.linspace(0, 1, num))
        cmap_name = base.name + str(num)
        return base.from_list(cmap_name, color_list, num)

    if actions is None:
        actions = td.get("action", None)
    # if batch_size greater than 0 , we need to select the first batch element
    if td.batch_size != torch.Size([]):
        td = td[0]
        actions = actions[0]

    num_agents = td["num_agents"]
    locs = td["locs"]
    cmap = discrete_cmap(num_agents, "rainbow")

    fig, ax = plt.subplots()

    # Add depot action = 0 to before first action and after last action
    actions = torch.cat(
        [
            torch.zeros(1, dtype=torch.int64),
            actions,
            torch.zeros(1, dtype=torch.int64),
        ]
    )

    # Make list of colors from matplotlib
    for i, loc in enumerate(locs):
        if i == 0:
            # depot
            marker = "s"
            color = "g"
            label = "Depot"
            markersize = 10
        else:
            # normal location
            marker = "o"
            color = "tab:blue"
            label = "Customers"
            markersize = 8
        if i > 1:
            label = ""

        ax.plot(
            loc[0],
            loc[1],
            color=color,
            marker=marker,
            markersize=markersize,
            label=label,
        )

    # Plot the actions in order
    agent_idx = 0
    for i in range(len(actions)):
        if actions[i] == 0:
            agent_idx += 1
        color = cmap(num_agents - agent_idx)

        from_node = actions[i]
        to_node = (
            actions[i + 1] if i < len(actions) - 1 else actions[0]
        )  # last goes back to depot
        from_loc = td["locs"][from_node]
        to_loc = td["locs"][to_node]
        ax.plot([from_loc[0], to_loc[0]], [from_loc[1], to_loc[1]], color=color)
        ax.annotate(
            "",
            xy=(to_loc[0], to_loc[1]),
            xytext=(from_loc[0], from_loc[1]),
            arrowprops=dict(arrowstyle="->", color=color),
            annotation_clip=False,
        )

    # Legend
    handles, labels = ax.get_legend_handles_labels()
    ax.legend(handles, labels)
    ax.set_title("mTSP")
    ax.set_xlabel("x-coordinate")
    ax.set_ylabel("y-coordinate")

    return ax



import os

from typing import Optional

import numpy as np
import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.pylogger import get_pylogger

from .generator import DPPGenerator

log = get_pylogger(__name__)


class DPPEnv(RL4COEnvBase):
    """Decap Placement Problem (DPP) as done in DevFormer paper: https://arxiv.org/abs/2205.13225

    The environment is a 10x10 grid with 100 locations containing either a probing port or a keepout region.
    The goal is to place decaps (decoupling capacitors) to maximize the impedance suppression at the probing port.
    Decaps cannot be placed in keepout regions or at the probing port and the number of decaps is limited.

    Observations:
        - locations of the probing port and keepout regions
        - current decap placement
        - remaining decaps

    Constraints:
        - decaps cannot be placed at the probing port or keepout regions
        - the number of decaps is limited

    Finish Condition:
        - the number of decaps exceeds the limit

    Reward:
        - the impedance suppression at the probing port

    Args:
        generator: DPPGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "dpp"

    def __init__(
        self,
        generator: DPPGenerator = None,
        generator_params: dict = {},
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = DPPGenerator(**generator_params)
        self.generator = generator

        self.max_decaps = self.generator.max_decaps
        self.size = self.generator.size
        self.raw_pdn = self.generator.raw_pdn
        self.decap = self.generator.decap
        self.freq = self.generator.freq
        self.num_freq = self.generator.num_freq
        self.data_dir = self.generator.data_dir

        self._make_spec(self.generator)

    def _step(self, td: TensorDict) -> TensorDict:
        current_node = td["action"]

        # Set available to 0 (i.e., already placed) if the current node is the first node
        available = td["action_mask"].scatter(
            -1, current_node.unsqueeze(-1).expand_as(td["action_mask"]), 0
        )

        # Set done if i is greater than max_decaps
        done = td["i"] >= self.max_decaps - 1

        # The reward is calculated outside via get_reward for efficiency, so we set it to 0 here
        reward = torch.zeros_like(done)

        td.update(
            {
                "i": td["i"] + 1,
                "action_mask": available,
                "reward": reward,
                "done": done,
            }
        )
        return td

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        device = td.device

        # Other variables
        i = torch.zeros((*batch_size, 1), dtype=torch.int64, device=device)

        return TensorDict(
            {
                "locs": td["locs"],
                "probe": td["probe"],
                "i": i,
                "action_mask": td["action_mask"],
                "keepout": ~td["action_mask"],
            },
            batch_size=batch_size,
        )

    def _make_spec(self, generator: DPPGenerator):
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(generator.size**2, 2),
                dtype=torch.float32,
            ),
            probe=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            keepout=Unbounded(
                shape=(generator.size**2),
                dtype=torch.bool,
            ),
            i=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=Unbounded(
                shape=(generator.size**2),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=generator.size**2,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    def _get_reward(self, td, actions):
        """
        We call the reward function with the final sequence of actions to get the reward
        Calling per-step would be very time consuming due to decap simulation
        """
        # We do the operation in a batch
        if len(td.batch_size) == 0:
            td = td.unsqueeze(0)
            actions = actions.unsqueeze(0)
        probes = td["probe"]
        reward = torch.stack(
            [self._decap_simulator(p, a) for p, a in zip(probes, actions)]
        )
        return reward

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor):
        assert True, "Not implemented"

    def _decap_placement(self, pi, probe):
        device = pi.device

        n = m = self.size  # columns and rows
        num_decap = torch.numel(pi)
        z1 = self.raw_pdn.to(device)

        decap = self.decap.reshape(-1).to(device)
        z2 = torch.zeros(
            (self.num_freq, num_decap, num_decap), dtype=torch.float32, device=device
        )

        qIndx = torch.arange(num_decap, device=device)

        z2[:, qIndx, qIndx] = torch.abs(decap)[:, None].repeat_interleave(
            z2[:, qIndx, qIndx].shape[-1], dim=-1
        )
        pIndx = pi.long()

        aIndx = torch.arange(len(z1[0]), device=device)
        aIndx = torch.tensor(
            list(set(aIndx.tolist()) - set(pIndx.tolist())), device=device
        )

        z1aa = z1[:, aIndx, :][:, :, aIndx]
        z1ap = z1[:, aIndx, :][:, :, pIndx]
        z1pa = z1[:, pIndx, :][:, :, aIndx]
        z1pp = z1[:, pIndx, :][:, :, pIndx]
        z2qq = z2[:, qIndx, :][:, :, qIndx]

        zout = z1aa - torch.matmul(torch.matmul(z1ap, torch.inverse(z1pp + z2qq)), z1pa)

        idx = torch.arange(n * m, device=device)
        mask = torch.zeros(n * m, device=device).bool()
        mask[pi] = True
        mask = mask & (idx < probe)
        probe -= mask.sum().item()

        zout = zout[:, probe, probe]
        return zout

    def _decap_model(self, z_initial, z_final):
        impedance_gap = torch.zeros(self.num_freq, device=self.device)

        impedance_gap = z_initial - z_final
        reward = torch.sum(impedance_gap * 1000000000 / self.freq.to(self.device))

        reward = reward / 10
        return reward

    def _initial_impedance(self, probe):
        zout = self.raw_pdn.to(self.device)[:, probe, probe]
        return zout

    def _decap_simulator(self, probe, solution, keepout=None):
        self.to(self.device)

        probe = probe.item()

        assert len(solution) == len(
            torch.unique(solution)
        ), "An Element of Decap Sequence must be Unique"

        if keepout is not None:
            keepout = torch.tensor(keepout)
            intersect = torch.tensor(list(set(solution.tolist()) & set(keepout.tolist())))
            assert len(intersect) == 0, "Decap must be not placed at the keepout region"

        z_initial = self._initial_impedance(probe)
        z_initial = torch.abs(z_initial)
        z_final = self._decap_placement(solution, probe)
        z_final = torch.abs(z_final)
        reward = self._decap_model(z_initial, z_final)
        return reward

    def _load_dpp_data(self, chip_file, decap_file, freq_file):
        def _load_file(fpath):
            f = os.path.join(self.generator.data_dir, fpath)
            if not os.path.isfile(f):
                self._download_data()
            with open(f, "rb") as f_:
                return torch.from_numpy(np.load(f_)).to(self.device)

        self.raw_pdn = _load_file(chip_file)  # [num_freq, size^2, size^2]
        self.decap = _load_file(decap_file).to(torch.complex64)  # [num_freq, 1, 1]
        self.freq = _load_file(freq_file)  # [num_freq]
        self.size = int(np.sqrt(self.raw_pdn.shape[-1]))
        self.num_freq = self.freq.shape[0]



import abc

from typing import Tuple

import torch.nn as nn

from tensordict import TensorDict
from torch import Tensor

from rl4co.envs import RL4COEnvBase
from rl4co.models.nn.env_embeddings import env_init_embedding
from rl4co.models.nn.pos_embeddings import pos_init_embedding
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class ImprovementEncoder(nn.Module):
    """Base class for the encoder of improvement models"""

    def __init__(
        self,
        embed_dim: int = 128,
        init_embedding: nn.Module = None,
        pos_embedding: nn.Module = None,
        env_name: str = "pdp_ruin_repair",
        pos_type: str = "CPE",
        num_heads: int = 4,
        num_layers: int = 3,
        normalization: str = "layer",
        feedforward_hidden: int = 128,
        linear_bias: bool = False,
    ):
        super(ImprovementEncoder, self).__init__()

        if isinstance(env_name, RL4COEnvBase):
            env_name = env_name.name
        self.env_name = env_name
        self.init_embedding = (
            env_init_embedding(
                self.env_name, {"embed_dim": embed_dim, "linear_bias": linear_bias}
            )
            if init_embedding is None
            else init_embedding
        )

        self.pos_type = pos_type
        self.pos_embedding = (
            pos_init_embedding(self.pos_type, {"embed_dim": embed_dim})
            if pos_embedding is None
            else pos_embedding
        )

    @abc.abstractmethod
    def _encoder_forward(self, init_h: Tensor, init_p: Tensor) -> Tuple[Tensor, Tensor]:
        """Process the node embeddings and positional embeddings to the final embeddings

        Args:
            init_h: initialized node embeddings
            init_p: initialized positional embeddings

        Returns:
            Tuple containing the final node embeddings and final positional embeddings (if any)
        """
        raise NotImplementedError("Implement me in subclass!")

    def forward(self, td: TensorDict) -> Tuple[Tensor, Tensor]:
        """Forward pass of the encoder.
        Transform the input TensorDict into a latent representation.

        Args:
            td: Input TensorDict containing the environment state

        Returns:
            h: Latent representation of the input
            init_h: Initial embedding of the input
        """
        # Transfer to embedding space (node)
        init_h = self.init_embedding(td)

        # Transfer to embedding space (solution)
        init_p = self.pos_embedding(td)

        # Process embedding
        final_h, final_p = self._encoder_forward(init_h, init_p)

        # Return latent representation and initial embedding
        return final_h, final_p


class ImprovementDecoder(nn.Module, metaclass=abc.ABCMeta):
    """Base decoder model for improvement models. The decoder is responsible for generating the logits of the action"""

    @abc.abstractmethod
    def forward(self, td: TensorDict, final_h: Tensor, final_p: Tensor) -> Tensor:
        """Obtain logits to perform operators that improve the current solution to the next ones

        Args:
            td: TensorDict with the current environment state
            final_h: final node embeddings
            final_p: final positional embeddings

        Returns:
            Tuple containing the logits
        """
        raise NotImplementedError("Implement me in subclass!")


class ImprovementPolicy(nn.Module):
    """
    Base class for improvement policies. Improvement policies take an instance + a solution as input and output a specific operator that changes the current solution to a new one.

    "Improvement" means that a solution is (potentially) improved to a new one by the model.

    """

    @abc.abstractmethod
    def forward(
        self,
        td: TensorDict,
        env: str | RL4COEnvBase = None,
        phase: str = "train",
        return_actions: bool = True,
        return_entropy: bool = False,
        return_init_embeds: bool = False,
        actions=None,
        **decoding_kwargs,
    ) -> dict:
        """Forward pass of the policy.

        Args:
            td: TensorDict containing the environment state
            env: Environment to use for decoding. If None, the environment is instantiated from `env_name`. Note that
                it is more efficient to pass an already instantiated environment each time for fine-grained control
            phase: Phase of the algorithm (train, val, test)
            return_actions: Whether to return the actions
            return_entropy: Whether to return the entropy
            return_init_embeds: Whether to return the initial embeddings
            actions: Actions to use for evaluating the policy.
                If passed, use these actions instead of sampling from the policy to calculate log likelihood
            decoding_kwargs: Keyword arguments for the decoding strategy. See :class:`rl4co.utils.decoding.DecodingStrategy` for more information.

        Returns:
            out: Dictionary containing the reward, log likelihood, and optionally the actions and entropy
        """
        raise NotImplementedError("Implement me in subclass!")
